/* ---------------------------------------------------------------------------
 *
 *   This file is part of OpenFTPD (based on FTP4ALL)
 *   (C) 1996-2000 by Crescent, (C) 2000 by the OpenFTPD Team
 *
 * ---------------------------------------------------------------------------
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 * ---------------------------------------------------------------------------
 *   MODULE: f4adp_user.c - functions to work with users
 * ---------------------------------------------------------------------------
 */

#include "ac_config.h"
#include <sys/types.h>
#include <ctype.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#ifdef CRYPT_CRYPT_H
#include <crypt.h>
#elif defined CRYPT_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#include "externsd.h"
#include "ftpd.h"
#include "group.h"
#include "serverd.h"
#include "user.h"
#include "utils.h"
#include "sections.h"
#include "log.h"
#include "f4adp_admin.h"
#include "strlcpy.h"
#include "strlcat.h"

int f4adp_gop_add(struct child_s* c, char** p, int n)
{ 
   struct user_s *u;

   u = get_user_by_name(p[0]);
   if (!u) { 
      server_reply(c,"-User does not exist");
      return 0;
   }
   u->rights |= groupop_rights;
   u->class = UC_GROUPOP;
   server_reply(c,"+Group operator status granted.");
   return 0;
}


int f4adp_gop_del(struct child_s* c, char** p, int n)
{ 
   struct user_s *u;

   u = get_user_by_name(p[0]);
   if (!u) { 
      server_reply(c,"-User does not exist");
      return 0;
   }
   u->rights &=~ groupop_rights;
   u->class = UC_NORMAL;
   server_reply(c,"+Group operator status revoked.");
   return 0;
}


int f4adp_op_add(struct child_s* c, char** p, int n)
{ 
   struct user_s *u;

   u = get_user_by_name(p[0]);
   if (!u) { 
      server_reply(c,"-User does not exist");
      return 0;
   }
   u->rights |= siteop_rights;
   u->class = UC_SITEOP;
   server_reply(c,"+Site operator status granted.");
   return 0;
}


int f4adp_op_del(struct child_s* c, char** p, int n)
{ 
   struct user_s *u;

   u = get_user_by_name(p[0]);
   if (!u) { 
      server_reply(c,"-User does not exist");
      return 0;
   }
   u->rights &=~ siteop_rights;
   u->class = UC_NORMAL;
   server_reply(c,"+Site operator status revoked.");
   return 0;
}


/* Look for free slots in a group */
int has_group_free_slots(int gid, char *str, int size)
{ 
   struct group_s *g= NULL;
   struct user_s  *u;
   int counted = 0;
   int i;

   g = get_group_by_gid(gid);
   if (!g) {
      g_snprintf(str, size, "-Group with gid %d does not exist", gid);
      return 1;
   }
   if (!g->max_slots) return 0;

   for (i = 0 ; i < user_count ; i++) {
      u=user+i;
      if(u->gid == g->gid && !templateuser(u) && !invisible(u) && 
         !(u->class & UF_DISABLED))
         ++counted;
   }
   if (counted >= g->max_slots ) {
      g_snprintf(str, size, "-Group %s has all the slots filled",g->name);
      return 1; 
   }
   return 0;
}

/* user add/copy */
int f4adp_usr_add(struct child_s* c, char** p, int n)
{ 
   struct group_s*  g = NULL;
   int              i;
   int              found;
   int              x;
   struct user_s*   u = NULL;
   struct user_s    usr;
   char             str[256];
   struct sstat_s   usr_stat;
   struct sstat_s   *new_stat;

   if (strlen(p[0]) >= sizeof(usr.name)) { 
      server_reply(c,"-Username too long.");
      return 0;
   }
   if (get_user_by_name(p[0])) { 
      server_reply(c,"-User already exists.");
      return 0;
   }
   if (n == 2) { 
      u = get_user_by_name(p[1]);
      if (!u) {
         if (!groupop(c->user)) g = get_group_by_name(p[1]);
         if (!g) {
            server_reply(c,"-Template does not exist.");
            return 0;
	 }
      }
   } 
   if (u) { 
      usr = *u;
      usr.class &= ~UC_MASK;
      for (i=0; i<sizeof(usr.gids)/sizeof(gint32); i++) {
         if (usr.gids[i] != -1) {
            if(has_group_free_slots(usr.gids[i], str, sizeof(str)) == 1) {
              server_reply(c, str);
              return 0;
            }
         }
      }
      if (has_group_free_slots(usr.gid, str, sizeof(str)) == 1) {
         server_reply(c, str);
         return 0;
      }
   } 
   else { 
      memset(&usr, 0, sizeof(usr));
      for (i = 0; i < sizeof(usr.gids) / sizeof(gint32); i++) usr.gids[i] =-1;
   }

   if (g || groupop(c->user)) { 
      if (!g) g = get_group_by_gid(c->user->gid);
      if (!g) { 
         server_reply(c,"-Group does not exist.");
         return 0;
      }
      if (has_group_free_slots(g->gid, str, sizeof(str)) == 1) {
         server_reply(c, str);
         return 0;
      } 
      usr.gid      = g->gid;
      usr.limit    = g->limit;
      usr.credit   = g->credit;
      usr.dl_speed = g->dl_speed;
      usr.ul_speed = g->ul_speed;
      usr.class    = g->logins << 8;
      usr.rights   = g->rights;
      strlcpy(usr.info, g->info, sizeof(usr.info));
      memmove(usr.idents[0], g->ident, sizeof(g->ident));
   } 
   else if (!u) { 
      usr.gid = -1;
      usr.class = 0x0100;
   }
   strlcpy(usr.name, p[0], sizeof(usr.name));
   usr.uid = get_new_uid();
   usr.class |= UC_NORMAL;
   strlcpy(usr.creator, c->user->name, sizeof(usr.creator));
   time(&usr.creation_time);

   memset(&usr_stat, 0, sizeof(struct sstat_s));
   usr_stat.uid = usr.uid;
   
   for (i = 0; i < section_count; i++) {
      found = 0;
      for (x = 0; x < i; x++) {
         if (!strcasecmp((section + x)->stat_file, (section + i)->stat_file)) {
               (section + i)->stat = (section + x)->stat;
               (section + i)->stat_size = (section + x)->stat_size;
               (section + i)->stat_count = (section + x)->stat_count;
               found = 1;
         }
      }
      if (!found) {
         new_stat = g_realloc((section+i)->stat, ((section+i)->stat_count + 1) * sizeof(struct sstat_s));
         if (new_stat) { 
            memmove(new_stat+(section+i)->stat_count, &usr_stat, sizeof(struct sstat_s));
            (section+i)->stat = new_stat;
  	    (section+i)->stat_count++;
 	    (section+i)->stat_size++;
            section_stat_dirty = 1;
         }
         else { 
            server_reply(c,"-Out of memory");
            return 0;      
         }
      }
   }
   if (autohome && homepath != NULL) {
      if (*(homepath + strlen(homepath) - 1) != '/')
         g_snprintf(usr.home, sizeof(usr.home), "%s/%s", homepath, usr.name);
      else
         g_snprintf(usr.home, sizeof(usr.home), "%s%s", homepath, usr.name);
   }
   if (!insert_user(&usr)) { 
      server_reply(c,"-Out of memory");
      return 0;
   }
   if (*usr.home) {/*buba*/
      if (create_home(usr.uid, usr.gid, usr.home, str, sizeof(str))) {
         server_reply(c, "+User added but homedir could'nt be created"); 
         server_log_local(c, LG_SECURITY, "USER %s was added by %s", usr.name, 
                          c->user->name);
         return 0; 
      }
   }
   server_reply(c,"+User added");

   server_log_local(c, LG_SECURITY, "USER %s was added by %s", usr.name, 
      c->user->name);
   return 0;
}


/* user add quick (SITE ADDUSER) */
int f4adp_usr_adq(struct child_s* c, char** p, int n)
{ 
   struct group_s*  g;
   int 	            i;
   int              found;
   int              x;
   char             salt[2];
   struct user_s    usr;
   char             str[256];
   struct sstat_s   usr_stat;
   struct sstat_s   *new_stat;
   int              gop = 0;

   if (!sscanf(p[0], "%i", &gop) || gop < 0 || gop > 1) {
      server_reply(c,"-Syntax error.");
      return 0;
   }
   if (strlen(p[1]) >= sizeof(usr.name)) { 
      server_reply(c,"-Username too long.");
      return 0;
   }
   if (get_user_by_name(p[1])) { 
      server_reply(c,"-User already exists.");
      return 0;
   }
   memset(&usr, 0, sizeof(usr));
   memset(usr.gids, -1, sizeof(usr.gids));
   strlcpy(usr.name, p[1], sizeof(usr.name));

   /* site gadduser, primary group of the groupop is used */
   if (gop || groupop(c->user)) g = get_group_by_gid(c->user->gid);
   else {
      g = get_group_by_name(p[2]);
   }        
   if (!g) { 
      server_reply(c,"-Group does not exist.");
      return 0;
   }

   if (groupop(c->user) && has_group_free_slots(g->gid, str, sizeof(str)) == 1) {
      server_reply(c, str);
      return 0;
   }

   usr.gid      = g->gid;
   usr.limit    = g->limit;
   usr.credit   = g->credit;
   usr.dl_speed = g->dl_speed;
   usr.ul_speed = g->ul_speed;
   usr.class    = g->logins << 8;
   usr.rights   = g->rights;
   strlcpy(usr.info, g->info, sizeof(usr.info));

   memmove(usr.idents[0], g->ident, sizeof(g->ident));
   /* Password */
   salt[0] = salt_a[time(NULL) & 0x3F];
   salt[1] = salt_a[(time(NULL) >> 6) & 0x3F];
   strlcpy(usr.password, (char *)crypt(p[3], salt), sizeof(usr.password));

   /* Limit , NOT changable by groupops, use group limit instead */
   if ( !groupop(c->user) ) { 
      if (!strcasecmp(p[4], "leech")) usr.limit = 0;
      else if (sscanf(p[4], "1:%i",&i)) usr.limit = 100 / i;
      else if (sscanf(p[4], "%i",&i)) usr.limit = i;
      else { 
         server_reply(c,"-Invalid limit.");
         return 0;
      }
   }
   /* IPs */
   for (i = 0; i < (n-5); i++) {
      if (get_iprange(p+5+i, usr.idents[i]) != 0) { 
	 server_reply(c,"-Invalid ident.");
	 return 0;
      }
   }
   /* Others */
   usr.uid = get_new_uid();
   usr.class |= UC_NORMAL;
   strlcpy(usr.creator, c->user->name, sizeof(usr.creator));
   time(&usr.creation_time);

   /* expand section stats */
   memset(&usr_stat, 0, sizeof(struct sstat_s));
   usr_stat.uid = usr.uid;

   for (i = 0; i < section_count; i++) {
      found = 0;
      for (x = 0; x < i; x++) {
         if (!strcasecmp((section + x)->stat_file, (section + i)->stat_file)) {
               (section + i)->stat = (section + x)->stat;
               (section + i)->stat_size = (section + x)->stat_size;
               (section + i)->stat_count = (section + x)->stat_count;
               found = 1;
         }
      }
      if (!found) {
         new_stat = g_realloc((section+i)->stat, ((section+i)->stat_count + 1) * sizeof(struct sstat_s));
         if (new_stat) { 
            memmove(new_stat+(section+i)->stat_count, &usr_stat, sizeof(struct sstat_s));
            (section+i)->stat = new_stat;
  	    (section+i)->stat_count++;
 	    (section+i)->stat_size++;
            section_stat_dirty = 1;
         }
         else { 
            server_reply(c,"-Out of memory");
            return 0;      
         }
      }
   }

   if (autohome && homepath != NULL) {
      if (*(homepath + strlen(homepath) - 1) != '/')
         g_snprintf(usr.home, sizeof(usr.home), "%s/%s", homepath, usr.name);
      else
         g_snprintf(usr.home, sizeof(usr.home), "%s%s", homepath, usr.name);
   }
   if (!insert_user(&usr)) { 
      server_reply(c,"-Out of memory.");
      return 0;
   }
   if (*usr.home) {/*buba*/
      if (create_home(usr.uid, usr.gid, usr.home, str, sizeof(str))) {
         server_reply(c, "+User added but homedir could'nt be created"); 
         server_log_local(c, LG_SECURITY, "USER %s was added by %s", usr.name, 
                          c->user->name);
         return 0; 
      }
   }
   server_log_local(c, LG_SECURITY, "USER %s was added by %s", usr.name, 
      c->user->name);

   server_reply(c,"+User added");
   return 0;
}


/* update user activity */
int f4adp_usr_activity(struct child_s* c, char** p, int n)
{
   int  action = 0;
   char str[1024];
 
   /* old: strlcpy(str, "", sizeof(str));*/
   /* new: memset(str, 0, sizeof(str)); */
   memset(str, 0, sizeof(str));

   sscanf(p[0], "%d", &c->hidden);
   sscanf(p[1], "%d", &action);
   sscanf(p[2], SI64, &(c->speed));
   if (n == 4)
     sscanf(p[3], "%1023c", str);

   switch (action) {
      case ACT_IDLE: 
         strlcpy(c->activity, "User is idle.", sizeof(c->activity));
	 memset(c->filename, 0, sizeof(c->filename));
	 /*  strlcpy(c->filename, "", sizeof(c->filename)); */
	 if (n == 4) 
            strlcpy(c->cwd, str, sizeof(c->cwd)); 
	 break;
	      
      case ACT_UPLOAD: 
         strlcpy(c->filename, strrchr(str, '/') ? strrchr(str, '/') + 1 : 
                 str, sizeof(c->filename));
	 /*
	 strlcpy(c->cwd, str,
		 strlen(str) - strlen(c->filename) < sizeof(c->cwd) ?
		 strlen(str) - strlen(c->filename) : 
                 sizeof(c->cwd));
	 */
	 strlcpy(c->cwd, str, sizeof(c->cwd));
         g_snprintf(c->activity, sizeof(c->activity), "UL: %s", 
		    c->filename);
	 break;

      case ACT_DOWNLOAD: 
         strlcpy(c->filename, strrchr(str, '/') ? strrchr(str, '/') + 1 : 
                 str, sizeof(c->filename));
	 strlcpy(c->cwd, str,
		 strlen(str) - strlen(c->filename) < sizeof(c->cwd) ?
		 strlen(str) - strlen(c->filename) : 
                 sizeof(c->cwd));
         g_snprintf(c->activity, sizeof(c->activity), "DL: %s", 
            c->filename);
	 break;

      case ACT_SITE_WHO: 
         strlcpy(c->activity, "Watching you!", sizeof(c->activity)); 
         strlcpy(c->filename, "", sizeof(c->filename));
	 if (n == 4) 
            strlcpy(c->cwd, str, sizeof(c->cwd)); 
         break;
   }

   server_reply(c,"+ok\n");
   return 0;
}


/* update user's current working dir */
int f4adp_usr_cwd(struct child_s* c, char** p, int n)
{
   sscanf(*p, "%d", &c->hidden);
   n--;
   p++; 
   memset(c->cwd, 0, sizeof(c->cwd));
   while (n--) { 
      if (strlen(c->cwd) + strlen(*p) + 1 < sizeof(c->cwd)) { 
         if (*c->cwd)
	   strlcat(c->cwd, " ", sizeof(c->cwd));
         strlcat(c->cwd, *p, sizeof(c->cwd));
      } 
      else { 
         if (strlen(c->cwd) < sizeof(c->cwd) - 3) { 
	    strlcat(c->cwd, " ", sizeof(c->cwd));
            strlcat(c->cwd, *p, sizeof(c->cwd));
         }
         n = 0;
      }
      p++;
   }
   server_reply(c,"+ok\n");
   return 0;
}


/* user change */
int f4adp_usr_chg(struct child_s* c, char** p, int n)
{ 
   struct child_s* ch;       /* children (aux)          */
   int             i;        /* integer (aux)           */
   gint32           j;        /* integer (aux)           */
   gint64           l;        /* integer (aux)           */
   int             pd;       /* flag: permission denied */
   int             rf;       /* range full (for ip+)    */
   char*           q;        /* argument after "="      */
   int             se;       /* flag: syntax error      */
   struct user_s*  u;        /* user to change          */
   struct user_s   usr;      /* user to change (temp)   */
   char            salt[2];  /* for PASSWD set        */
   struct group_s* g;
   char            str[128]; /* Aux buffer */
   struct sstat_s *st;
   gint64           r;

   u = get_user_by_name(p[0]);
   if (!u) { 
      server_reply(c,"-User does not exist");
      return 0;
   }
   pd = !allowed(c->user, R_USER_EDIT, u);
   if (pd) n = 1; else usr = *u;
   se = 0;
   rf = 0;
   while (--n) { 
      p++;
      q = strchr(*p,'=');
      if (q) { 
         *(q++) = '\0';

         /* NAME */
	 if (!strcasecmp(*p, "NAME")) { 
	    if (strlen(q) >= sizeof(usr.name)) se = 1; 
	    else { 
	       if (get_user_by_name(q)) { 
	          server_reply(c,"-User already exists");
                  return 0;
               }
               strlcpy(usr.name, q, sizeof(usr.name));
            }
	 } 

         /* UID */
	 else if (!strcasecmp(*p,"UID"))	{ 
	    if (!allowed(c->user,R_USER_EDIT_UID,u)) pd = 1; 
	    else if (!sscanf(q,SI32,&i)) se = 1;
            else {
               if (get_user_by_uid(i)) {
         	  server_reply(c,"-User already exists");
               }
               for (j = 0; j < section_count; j++) {
        	  if (section+j) {
                     st = get_stat_by_uid(section+j, usr.uid);
                     if (st) st->uid = i;
        	  }
               } 
               usr.uid = i;
            }
	 }

         /* GID */
	 else if (!strcasecmp(*p, "GID")) { 
	    if (!allowed(c->user,R_USER_EDIT_GRPS,u)) pd = 1;
            else {
               if (!sscanf(q,SI32,&i)) se = 1;
               else {
                  if (groupop(c->user) && (!groupopof(c->user, usr.gid) || 
		     !groupopof(c->user,i))) pd = 1;
                  else {
        	     if (has_group_free_slots(i, str, sizeof(str)) == 1) {
                	server_reply(c,str);
                	return 0;
        	     }
        	     usr.gid = i;
                  }
               }
            }
	 }

         /* GROUP */
	 else if (!strcasecmp(*p, "GROUP")) {
            if (!allowed(c->user,R_USER_EDIT_GRPS,u)) pd = 1;
            else {
               g = get_group_by_name(q);
               if (!g) {
                  server_reply(c,"-Group does not exist");
                  return 0;
               }
               if (groupop(c->user) && (!groupopof(c->user,usr.gid) || 
	           !groupopof(c->user,g->gid))) pd = 1;
               else {
                  if (has_group_free_slots(g->gid, str, sizeof(str)) == 1) {
        	     server_reply(c,str);
        	     return 0;
                  } 
		  else usr.gid = g->gid;
               }
            }
	 }

	 /* RIGHTS */
	 else if (!strcasecmp(*p, "RIGHTS")) { 
	    if (!allowed(c->user, R_USER_EDIT_RIGHTS, u)) pd = 1; 
	    else if (sscanf(q, SI64, &r)) usr.rights = r;
	    else se = 1;
	 } 

	 /* RIGHTS+ */
	 else if (!strcasecmp(*p,"RIGHTS+")) {             
            if (!allowed(c->user, R_USER_EDIT_RIGHTS, u)) pd =1;
            else {
               if (strcasecmp(q, "R_USER_ALL") == 0) {
                  usr.rights |= R_USER_ALL;
               } 
	       else if (strcasecmp(q, "R_ALWAYS_LOGIN") == 0) {
		  usr.rights |= R_ALWAYS_LOGIN;
	       }
	       else if (strcasecmp(q, "R_USER_DISABLE") == 0) {
                  usr.rights |= R_USER_DISABLE;
               } 
               else if (strcasecmp(q, "R_USER_LOCK") == 0) {
                  usr.rights |= R_USER_LOCK;
               }
	       else if (strcasecmp(q, "R_USER_KICK") == 0) {
                  usr.rights |= R_USER_KICK;
               } 
	       else if (strcasecmp(q, "R_USER_PASSWD") == 0) {
                  usr.rights |= R_USER_PASSWD;
               } 
	       else if (strcasecmp(q, "R_USER_ADD") == 0) {
                  usr.rights |= R_USER_ADD;
               } 
	       else if (strcasecmp(q, "R_USER_EDIT") == 0) {
                  usr.rights |= R_USER_EDIT;
               } 
	       else if (strcasecmp(q, "R_USER_DELETE") == 0) { 
                  usr.rights |= R_USER_DELETE;
               } 
	       else if (strcasecmp(q, "R_USER_READ") == 0) {
                  usr.rights |= R_USER_READ;
               } 
	       else if (strcasecmp(q, "R_USER_ADD_SO") == 0) {
                  usr.rights |= R_USER_ADD_SO;
               } 
	       else if (strcasecmp(q, "R_USER_EDIT_SO") == 0) {
                  usr.rights |= R_USER_EDIT_SO;
               } 
	       else if (strcasecmp(q, "R_USER_DELETE_SO") == 0) {
                  usr.rights |= R_USER_DELETE_SO;
               } 
	       else if (strcasecmp(q, "R_USER_READ_SO") == 0) {
                  usr.rights |= R_USER_READ_SO;
               } 
	       else if (strcasecmp(q,"R_USER_ADD_GO") == 0) {
                  usr.rights |= R_USER_ADD_GO;
               } 
	       else if (strcasecmp(q, "R_USER_EDIT_GO") == 0) {
                  usr.rights |= R_USER_EDIT_GO;
               } 
	       else if (strcasecmp(q, "R_USER_DELETE_GO") == 0) {
                  usr.rights |= R_USER_DELETE_GO;
               } 
	       else if (strcasecmp(q, "R_USER_READ_GO") == 0) {
                  usr.rights |= R_USER_READ_GO;
               } 
	       else if (strcasecmp(q, "R_USER_ADD_ANON") == 0) {
                  usr.rights |=R_USER_ADD_ANON;
               } 
	       else if (strcasecmp(q, "R_USER_ADD_TMP") == 0) {
                  usr.rights |=R_USER_ADD_TMP;
               }
	       else if (strcasecmp(q,"R_USER_EDIT_IP") == 0) {
                  usr.rights |= R_USER_EDIT_IP;
               } 
	       else if (strcasecmp(q, "R_USER_EDIT_GRPS") == 0) {
                  usr.rights |= R_USER_EDIT_GRPS;
               } 
	       else if (strcasecmp(q, "R_USER_EDIT_CREDIT") == 0) {
                  usr.rights |= R_USER_EDIT_CREDIT;
               } 
	       else if (strcasecmp(q, "R_USER_EDIT_STATS") == 0) {
                  usr.rights |= R_USER_EDIT_STATS;
               } 
	       else if (strcasecmp(q, "R_USER_EDIT_RIGHTS") == 0) {
                  usr.rights |= R_USER_EDIT_RIGHTS; 
               } 
	       else if (strcasecmp(q, "R_USER_EDIT_INFO") == 0) {
                  usr.rights |= R_USER_EDIT_INFO;
               } 
	       else if (strcasecmp(q, "R_USER_EDIT_UID") == 0) {
                  usr.rights |= R_USER_EDIT_UID;
               }
	       else if (strcasecmp(q, "R_USER_EDIT_LOGINS") == 0) {
                  usr.rights |= R_USER_EDIT_LOGINS;
               } 
	       else if (strcasecmp(q, "R_USER_EDIT_SPEEDS") == 0) {
                  usr.rights |= R_USER_EDIT_SPEEDS;
               } 
	       else if (strcasecmp(q, "R_ALL") == 0) {
                  usr.rights |= R_ALL;
               } 
	       else if (strcasecmp(q, "R_GROUP_ALL") == 0) {
                  usr.rights |= R_GROUP_ALL;
               } 
	       else if (strcasecmp(q, "R_GROUP_ADD") == 0) {
                  usr.rights |= R_GROUP_ADD;
               } 
	       else if (strcasecmp(q, "R_GROUP_EDIT") == 0) {
                  usr.rights |= R_GROUP_EDIT;
               } 
	       else if (strcasecmp(q, "R_GROUP_DELETE") == 0) {
                  usr.rights |= R_GROUP_DELETE;
               } 
	       else if (strcasecmp(q, "R_GROUP_LIST") == 0) { 
		  usr.rights |= R_GROUP_LIST;
               } 
	       else if (strcasecmp(q, "R_NUKE") == 0) { 
		  usr.rights |= R_NUKE;
               } 
	       else if (strcasecmp(q, "R_UNDUPE") == 0) { 
		  usr.rights |= R_UNDUPE;
               }
               else if (strcasecmp(q,"R_STATE") == 0) {
                  usr.rights |= R_STATE;
               }
               else if (strcasecmp(q,"R_PORTS") == 0) {
                  usr.rights |= R_PORTS;
               }
               else if (strcasecmp(q,"R_GUI") == 0) {
                  usr.rights |= R_GUI;
               }
               else if (strcasecmp(q,"R_MONITOR") == 0) {
                  usr.rights |= R_MONITOR;
               }
               else if (strcasecmp(q,"R_FILEACCESS") == 0) {
                  usr.rights |= R_FILEACCESS;
               }
               else if (strcasecmp(q,"R_CONFIG") == 0) {
                  usr.rights |= R_CONFIG;
               }
               else if (strcasecmp(q,"R_DUMP_FILES") == 0) {
                  usr.rights |= R_DUMP_FILES;
               }
               else if (strcasecmp(q,"R_REGISTRY") == 0) { 
                  usr.rights |= R_REGISTRY;
               }
               else if (strcasecmp(q,"R_MESSAGES") == 0) {
                  usr.rights |= R_MESSAGES;
               }
               else { 
		  se = 1; 
	       }
            }
	 }
   
         /* RIGHTS- */
	 else if (!strcasecmp(*p, "RIGHTS-")) {
            if (!allowed(c->user, R_USER_EDIT_RIGHTS, u)) pd =1;
            else {
               if (strcasecmp(q, "R_USER_ALL") == 0) {
                  usr.rights &=~ R_USER_ALL;
               } 
	       else if (strcasecmp(q, "R_ALWAYS_LOGIN") == 0) {
		  usr.rights &=~R_ALWAYS_LOGIN;
	       }
	       else if (strcasecmp(q, "R_USER_DISABLE") == 0) {
                  usr.rights &=~R_USER_DISABLE;
               }
               else if (strcasecmp(q, "R_USER_LOCK") == 0) {
                  usr.rights &=~R_USER_LOCK;
               }
	       else if (strcasecmp(q, "R_USER_KICK") == 0) {
                  usr.rights &=~R_USER_KICK;
               }
	       else if (strcasecmp(q, "R_USER_PASSWD") == 0) {
                  usr.rights &=~R_USER_PASSWD;
               }
	       else if (strcasecmp(q, "R_USER_ADD") == 0) {
                  usr.rights &=~R_USER_ADD;
               }
	       else if (strcasecmp(q, "R_USER_EDIT") == 0) {
                  usr.rights &=~R_USER_EDIT;
               }
	       else if (strcasecmp(q, "R_USER_DELETE") == 0) {
                  usr.rights &=~R_USER_DELETE;
               }
	       else if (strcasecmp(q, "R_USER_READ") == 0) {
                  usr.rights &=~R_USER_READ;
               }
	       else if (strcasecmp(q, "R_USER_ADD_SO") == 0) {
                  usr.rights &=~R_USER_ADD_SO;
               }
	       else if (strcasecmp(q, "R_USER_EDIT_SO") == 0) {
                  usr.rights &=~R_USER_EDIT_SO;
               }
	       else if (strcasecmp(q, "R_USER_DELETE_SO") == 0) {
                  usr.rights &=~R_USER_DELETE_SO;
               }
	       else if (strcasecmp(q, "R_USER_READ_SO") == 0) {
                  usr.rights &=~R_USER_READ_SO;
               }
	       else if (strcasecmp(q, "R_USER_ADD_GO") == 0) {
                  usr.rights &=~R_USER_ADD_GO;
               }
	       else if (strcasecmp(q, "R_USER_EDIT_GO") == 0) {
                  usr.rights &=~R_USER_EDIT_GO;
               }
	       else if (strcasecmp(q, "R_USER_DELETE_GO") == 0) {
                  usr.rights &=~R_USER_DELETE_GO;
               }
	       else if (strcasecmp(q, "R_USER_READ_GO") == 0) {
                  usr.rights &=~R_USER_READ_GO;
               }
	       else if (strcasecmp(q, "R_USER_ADD_ANON") == 0) {
                  usr.rights &=~R_USER_ADD_ANON;
               }
	       else if (strcasecmp(q, "R_USER_ADD_TMP") == 0) {
                  usr.rights &=~R_USER_ADD_TMP;
               }
	       else if (strcasecmp(q, "R_USER_EDIT_IP") == 0) {
                  usr.rights &=~R_USER_EDIT_IP;
               }
	       else if (strcasecmp(q, "R_USER_EDIT_GRPS") == 0) {
                  usr.rights &=~R_USER_EDIT_GRPS;
               }
	       else if (strcasecmp(q, "R_USER_EDIT_CREDIT") == 0) {
                  usr.rights &=~R_USER_EDIT_CREDIT;
               }
	       else if (strcasecmp(q, "R_USER_EDIT_STATS") == 0) {
                  usr.rights &=~R_USER_EDIT_STATS;
               }
	       else if (strcasecmp(q, "R_USER_EDIT_RIGHTS") == 0) {
                  usr.rights &=~R_USER_EDIT_RIGHTS;
               }
	       else if (strcasecmp(q, "R_USER_EDIT_INFO") == 0) {
                  usr.rights &=~R_USER_EDIT_INFO;
               }
	       else if (strcasecmp(q, "R_USER_EDIT_UID") == 0) {
                  usr.rights &=~R_USER_EDIT_UID;
               }
	       else if (strcasecmp(q, "R_USER_EDIT_LOGINS") == 0) {
                  usr.rights &=~R_USER_EDIT_LOGINS;
               }
	       else if (strcasecmp(q, "R_USER_EDIT_SPEEDS") == 0) {
                  usr.rights &=~R_USER_EDIT_SPEEDS;
               }
	       else if (strcasecmp(q, "R_ALL") == 0) {
                  usr.rights &=~R_ALL;
               }
	       else if (strcasecmp(q, "R_GROUP_ALL") == 0) {
                  usr.rights &=~R_GROUP_ALL;
               }
	       else if (strcasecmp(q, "R_GROUP_ADD") == 0) {
                  usr.rights &=~R_GROUP_ADD;
               }
	       else if (strcasecmp(q, "R_GROUP_EDIT") == 0) {
                  usr.rights &=~R_GROUP_EDIT;
               }
	       else if (strcasecmp(q, "R_GROUP_DELETE") == 0) {
                  usr.rights &=~R_GROUP_DELETE;
               }
	       else if (strcasecmp(q, "R_GROUP_LIST") == 0) {
                  usr.rights &=~R_GROUP_LIST;
               }
	       else if (strcasecmp(q, "R_NUKE") == 0) { 
		  usr.rights &=~R_NUKE;
               }
	       else if (strcasecmp(q, "R_UNDUPE") == 0) { 
		  usr.rights &=~R_UNDUPE;
               }
               else if (strcasecmp(q,"R_STATE") == 0) {
                  usr.rights &=~R_STATE;
               }
               else if (strcasecmp(q,"R_PORTS") == 0) {
                  usr.rights &=~R_PORTS;
               }
               else if (strcasecmp(q,"R_GUI") == 0) {
                  usr.rights &=~R_GUI;
               }
               else if (strcasecmp(q,"R_MONITOR") == 0) {
                  usr.rights &=~R_MONITOR;
               }
               else if (strcasecmp(q,"R_FILEACCESS") == 0) {
                  usr.rights &=~R_FILEACCESS;
               }
               else if (strcasecmp(q,"R_CONFIG") == 0) {
                  usr.rights &=~R_CONFIG;
               }
               else if (strcasecmp(q,"R_DUMP_FILES") == 0) {
                  usr.rights &=~R_DUMP_FILES;
               }
               else if (strcasecmp(q,"R_REGISTRY") == 0) {
                  usr.rights &=~R_REGISTRY;
               }
               else if (strcasecmp(q,"R_MESSAGES") == 0) {
                  usr.rights &=~R_MESSAGES;
               }
               else se = 1;
            }
	 }

         /* IPRANGE */
	 else if (!strcasecmp(*p, "IPRANGE") || !strcasecmp(*p, "IP0")) { 
	    if (!allowed(c->user, R_USER_EDIT_IP, u)) pd = 1; 
	    else if (get_iprange(&q, usr.idents[0]) != 0) se = 1;
	 } 
	 else if (!strcasecmp(*p,"IP1")) { 
	    if (!allowed(c->user, R_USER_EDIT_IP, u)) pd = 1; 
	    else if (get_iprange(&q, usr.idents[1]) != 0) se = 1;
	 } 
	 else if (!strcasecmp(*p, "IP2")) { 
	    if (!allowed(c->user, R_USER_EDIT_IP, u)) pd = 1; 
	    else if (get_iprange(&q, usr.idents[2]) != 0) se = 1;
	 } 
	 else if (!strcasecmp(*p, "IP3")) { 
	    if (!allowed(c->user, R_USER_EDIT_IP, u)) pd = 1; 
	    else if (get_iprange(&q, usr.idents[3]) != 0) se = 1;
	 } 
	 else if (!strcasecmp(*p, "IP4")) { 
	    if (!allowed(c->user, R_USER_EDIT_IP, u)) pd = 1; 
	    else if (get_iprange(&q, usr.idents[4]) != 0) se = 1;
	 } 
	 else if (!strcasecmp(*p, "IP5")) { 
	    if (!allowed(c->user, R_USER_EDIT_IP, u)) pd = 1; 
	    else if (get_iprange(&q, usr.idents[5]) != 0) se = 1;
	 } 
	 else if (!strcasecmp(*p, "IP6")) { 
	    if (!allowed(c->user, R_USER_EDIT_IP, u)) pd = 1; 
	    else if (get_iprange(&q, usr.idents[6]) != 0) se = 1;
	 } 
	 else if (!strcasecmp(*p, "IP7")) { 
	    if (!allowed(c->user, R_USER_EDIT_IP, u)) pd = 1; 
	    else if (get_iprange(&q, usr.idents[7]) != 0) se = 1;
	 } 
	 else if (!strcasecmp(*p, "IP8")) { 
	    if (!allowed(c->user, R_USER_EDIT_IP, u)) pd = 1; 
	    else if (get_iprange(&q, usr.idents[8]) != 0) se = 1;
	 } 
	 else if (!strcasecmp(*p, "IP9")) { 
	    if (!allowed(c->user, R_USER_EDIT_IP, u)) pd = 1; 
	    else if (get_iprange(&q, usr.idents[9]) != 0) se = 1;
	 } 
	 else if (!strcasecmp(*p, "IP+")) { 
	    if (!allowed(c->user, R_USER_EDIT_IP, u)) pd = 1; 
	    else for (i=0; i<10; i++) { 
	       if (!memcmp(usr.idents[i] + 12, "\0\0\0\0\0\0\0\0", 8)) { 
	          if (get_iprange(&q, usr.idents[i]) != 0) se = 1;
                  i = 10; rf = 0;
               } 
	       else rf = 1;
            } 
	 } 
	 else if (!strcasecmp(*p, "IP-")) { 
	    if (!allowed(c->user, R_USER_EDIT_IP, u)) pd = 1; 
	    else if (sscanf(q, SI32, &i)) { 
	       if ((i >= 0) && (i <= 9)) { 
	          memset(usr.idents[i], 0, sizeof(usr.idents[0]));
               } 
	       else se = 1;
            } 
            else se = 1;
	 } 
	 
	 else if (!strcasecmp(*p, "LIMIT")) { 
	    if (!allowed(c->user,R_USER_EDIT_CREDIT,u)) pd = 1; 
	    else if (!sscanf(q, SI32, &usr.limit)) se = 1;
	 } 

	 else if (!strcasecmp(*p, "INFO")) { 
	    if (!allowed(c->user,R_USER_EDIT_INFO,u)) pd = 1; 
	    else if (strlen(q) < sizeof(usr.info)) {
               strlcpy(usr.info, q, sizeof(usr.info)); 
               /* take out : and \t as are used as delimiters */
               for (i = 0; i < strlen(usr.info); i++) {
                   if (usr.info[i] == ':' || usr.info[i] == '\t') 
                      usr.info[i] = ' ';
               }
            }
	    else se = 1;
	 } 
         /* HOME */
         else if (!strcasecmp(*p, "HOME")) {
            if (!allowed(c->user, R_USER_ADD, u)) pd = 1;
            else if (strlen(q) < sizeof(usr.home)) {
               if (strlen(q) == 0) {
	          server_reply(c, "+Home directory reset");
		  strlcpy(usr.home, q, sizeof(usr.home));
		  return 0;
	       }
	       if (*q != '/') {
                  strlcpy(str, "-Home directory must begin with /", sizeof(str));
                  server_reply(c, str);
                  return 0;
               }
               strlcpy(usr.home, q, sizeof(usr.home));
               if (create_home(usr.uid, usr.gid, usr.home, str, sizeof(str))) {
                  server_reply(c, str); 
                  return 0; 
               }
            }
            else se = 1;
         }
            
         /* WEEKLY ALOTMENT */
         else if (!strcasecmp(*p, "WKALOT")) {
            if (!allowed(c->user, R_USER_EDIT_CREDIT, u)) pd = 1;
            else if (sscanf(q, SI64, &usr.wkly_alot)) 
                 usr.credit = usr.wkly_alot;
            else se = 1;
         }

         else if (!strcasecmp(*p, "MBWKALOT")) {
            if (!allowed(c->user, R_USER_EDIT_CREDIT, u)) pd = 1;
            else if (sscanf(q, SI64, &l)) {
               usr.wkly_alot = l * 0x100000;
               usr.credit = usr.wkly_alot;
            }
            else se = 1;
         }
         
	 /* CREDITS */
	 else if (!strcasecmp(*p, "CREDIT")) { 
	    if (!allowed(c->user, R_USER_EDIT_CREDIT, u)) pd = 1; 
	    else if (!sscanf(q, SI64, &usr.credit)) se = 1;
	 } 
	 
	 else if (!strcasecmp(*p, "CREDIT+")) { 
	    if (!allowed(c->user, R_USER_EDIT_CREDIT, u)) pd = 1; 
	    else if (sscanf(q, SI64, &l)) usr.credit += l; 
	    else se = 1;
	 } 
	 
	 else if (!strcasecmp(*p, "CREDIT-")) { 
	    if (!allowed(c->user, R_USER_EDIT_CREDIT, u)) pd = 1; 
	    else if (sscanf(q, SI64, &l)) usr.credit -= l; 
	    else se = 1;
	 } 
	 
	 else if (!strcasecmp(*p, "MBCRED")) { 
	    if (!allowed(c->user, R_USER_EDIT_CREDIT, u)) pd = 1; 
	    else if (sscanf(q, SI64, &l)) usr.credit = l * 0x100000; 
	    else se = 1;
	 } 
	 
	 else if (!strcasecmp(*p, "MBCRED+")) { 
	    if (!allowed(c->user, R_USER_EDIT_CREDIT, u)) pd = 1; 
	    else if (sscanf(q,SI64,&l)) usr.credit += l * 0x100000; 
	    else se = 1;
	 } 
	 
	 else if (!strcasecmp(*p, "MBCRED-")) { 
	    if (!allowed(c->user, R_USER_EDIT_CREDIT, u)) pd = 1; 
	    else if (sscanf(q,SI64,&l)) usr.credit -= l * 0x100000; 
	    else se = 1;
	 }
         /* STATS */
	 else if(!strcasecmp(*p, "MBUP+")) {  
	    if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
	    else if (sscanf(q, SI64, &l)) usr.ul_total+=l * 0x100000; 
	    else se = 1;
	 }  
	 else if(!strcasecmp(*p, "MBUP-")) {  
	    if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; 
	    else if (sscanf(q,SI64,&l)) usr.ul_total-=l * 0x100000; 
	    else se = 1;
	 } else if(!strcasecmp(*p,"MBUP"))
	 {  if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
           if (sscanf(q,SI64,&l)) usr.ul_total=l * 0x100000; else se = 1;
	 } else if (!strcasecmp(*p,"MBDOWN+"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_total+=l * 0x100000; else se = 1;
	 } else if (!strcasecmp(*p,"MBDOWN-"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_total-=l * 0x100000; else se = 1;
	 } else if (!strcasecmp(*p,"MBDOWN"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_total=l * 0x100000; else se = 1;
	 } else if (!strcasecmp(*p,"WKUP-"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.ul_week-=l * 0x100000; else se = 1;
	 } else if (!strcasecmp(*p,"WKDOWN-"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_week-=l * 0x100000; else se = 1;
	 } else if (!strcasecmp(*p,"WKUP+"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.ul_week+=l * 0x100000; else se = 1;
	 } else if (!strcasecmp(*p,"WKDOWN+"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_week+=l * 0x100000; else se = 1;
	 } else if (!strcasecmp(*p,"WKUP"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.ul_week=l * 0x100000; else se = 1;
	 } else if (!strcasecmp(*p,"WKDOWN"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_week=l * 0x100000; else se = 1;
	 } else if (!strcasecmp(*p,"MUP-"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.ul_month-=l * 0x100000; else se = 1;
	 } else if (!strcasecmp(*p,"MUP+"))   
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.ul_month+=l * 0x100000; else se = 1;
	 } else if (!strcasecmp(*p,"MUP"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.ul_month=l * 0x100000; else se = 1;
	 } else if (!strcasecmp(*p,"MDOWN+"))   
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_month+=l * 0x100000; else se = 1;
	 } else if (!strcasecmp(*p,"MDOWN-"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_month-=l * 0x100000; else se = 1;
	 } else if (!strcasecmp(*p,"MDOWN"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_month=l * 0x100000; else se = 1;
	 } else if (!strcasecmp(*p,"YUP-"))   
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.ul_year-=l * 0x100000; else se = 1;
	 } else if (!strcasecmp(*p,"YUP+"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.ul_year+=l * 0x100000; else se = 1;
	 } else if (!strcasecmp(*p,"YUP"))   
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.ul_year=l * 0x100000; else se = 1;
	 } else if (!strcasecmp(*p,"YDOWN+"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_year+=l * 0x100000; else se = 1;
	 } else if (!strcasecmp(*p,"YDOWN-"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_year-=l * 0x100000; else se = 1;
	 } else if (!strcasecmp(*p,"YDOWN"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_year=l * 0x100000; else se = 1;
	 } else if (!strcasecmp(*p,"DUP-"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.ul_day-=l * 0x100000; else se = 1;
	 } else if (!strcasecmp(*p,"DUP+"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.ul_day+=l * 0x100000; else se = 1;
	 } else if (!strcasecmp(*p,"DUP"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.ul_day=l * 0x100000; else se = 1;
	 } else if (!strcasecmp(*p,"DDOWN-"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_day-=l * 0x100000; else se = 1;
	 } else if (!strcasecmp(*p,"DDOWN+"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_day+=l * 0x100000; else se = 1;
	 } else if (!strcasecmp(*p,"DDOWN"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_day=l * 0x100000; else se = 1;
	 }
	 else if(!strcasecmp(*p,"KBUP+"))
	 {  if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
           if (sscanf(q,SI64,&l)) usr.ul_total+=l * 1024; else se = 1;
	 } else if(!strcasecmp(*p,"KBUP-"))
	 {  if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
           if (sscanf(q,SI64,&l)) usr.ul_total-=l * 1024; else se = 1;
	 } else if(!strcasecmp(*p,"KBUP"))
	 {  if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
           if (sscanf(q,SI64,&l)) usr.ul_total=l * 1024; else se = 1;
	 } else if (!strcasecmp(*p,"KBDOWN+"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_total+=l * 1024; else se = 1;
	 } else if (!strcasecmp(*p,"KBDOWN-"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_total-=l * 1024; else se = 1;
	 } else if (!strcasecmp(*p,"KBDOWN"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_total=l * 1024; else se = 1;
	 } else if (!strcasecmp(*p,"KBWKUP-"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.ul_week-=l * 1024; else se = 1;
	 } else if (!strcasecmp(*p,"KBWKDOWN-"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_week-=l * 1024; else se = 1;
	 } else if (!strcasecmp(*p,"KBWKUP+"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.ul_week+=l * 1024; else se = 1;
	 } else if (!strcasecmp(*p,"KBWKDOWN+"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_week+=l * 1024; else se = 1;
	 } else if (!strcasecmp(*p,"KBWKUP"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.ul_week=l * 1024; else se = 1;
	 } else if (!strcasecmp(*p,"KBWKDOWN"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_week=l * 1024; else se = 1;
	 } else if (!strcasecmp(*p,"KBMUP-"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.ul_month-=l * 1024; else se = 1;
	 } else if (!strcasecmp(*p,"KBMUP+"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.ul_month+=l * 1024; else se = 1;
	 } else if (!strcasecmp(*p,"KBMUP"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.ul_month=l * 1024; else se = 1;
	 } else if (!strcasecmp(*p,"KBMDOWN+"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_month+=l * 1024; else se = 1;
	 } else if (!strcasecmp(*p,"KBMDOWN-"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_month-=l * 1024; else se = 1;
	 } else if (!strcasecmp(*p,"KBMDOWN"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_month=l * 1024; else se = 1;
	 } else if (!strcasecmp(*p,"KBYUP-"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.ul_year-=l * 1024; else se = 1;
	 } else if (!strcasecmp(*p,"KBYUP+"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.ul_year+=l * 1024; else se = 1;
	 } else if (!strcasecmp(*p,"KBYUP"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.ul_year=l * 1024; else se = 1;
	 } else if (!strcasecmp(*p,"KBYDOWN+"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_year+=l * 1024; else se = 1;
	 } else if (!strcasecmp(*p,"KBYDOWN-"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_year-=l * 1024; else se = 1;
	 } else if (!strcasecmp(*p,"KBYDOWN"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_year=l * 1024; else se = 1;
	 } else if (!strcasecmp(*p,"KBDUP-"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.ul_day-=l * 1024; else se = 1;
	 } else if (!strcasecmp(*p,"KBDUP+"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.ul_day+=l * 1024; else se = 1;
	 } else if (!strcasecmp(*p,"KBDUP"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.ul_day=l * 1024; else se = 1;
	 } else if (!strcasecmp(*p,"KBDDOWN-"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_day-=l * 1024; else se = 1;
	 } else if (!strcasecmp(*p,"KBDDOWN+"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_day+=l * 1024; else se = 1;
	 } else if (!strcasecmp(*p,"KBDDOWN"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI64,&l)) usr.dl_day=l * 1024; else se = 1;
	 } 
	 else if (!strcasecmp(*p,"FUP+"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.ul_f_total+=i; else se = 1;
	 } else if (!strcasecmp(*p,"FUP-"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.ul_f_total-=i; else se = 1;
	 } else if (!strcasecmp(*p,"FUP"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.ul_f_total=i; else se = 1;
	 } else if (!strcasecmp(*p,"FDOWN-"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.dl_f_total-=i; else se = 1;
	 } else if (!strcasecmp(*p,"FDOWN"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.dl_f_total=i; else se = 1;
	 } else if (!strcasecmp(*p,"FDOWN+"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.dl_f_total+=i; else se = 1;
	 } else if (!strcasecmp(*p,"FWKUP+"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.ul_f_week+=i; else se = 1;
	 } else if (!strcasecmp(*p,"FWKUP-"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.ul_f_week-=i; else se = 1;
	 } else if (!strcasecmp(*p,"FWKUP"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.ul_f_week=i; else se = 1;
	 } else if (!strcasecmp(*p,"FWKDOWN-"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.dl_f_week-=i; else se = 1;
	 } else if (!strcasecmp(*p,"FWKDOWN"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.dl_f_week=i; else se = 1;
	 } else if (!strcasecmp(*p,"FWKDOWN+"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.dl_f_week+=i; else se = 1;
	 } else if (!strcasecmp(*p,"FMUP+"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.ul_f_month+=i; else se = 1;
	 } else if (!strcasecmp(*p,"FMUP-"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.ul_f_month-=i; else se = 1;
	 } else if (!strcasecmp(*p,"FMUP"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.ul_f_month=i; else se = 1;
	 } else if (!strcasecmp(*p,"FMDOWN-"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.dl_f_month-=i; else se = 1;
	 } else if (!strcasecmp(*p,"FMDOWN"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.dl_f_month=i; else se = 1;
	 } else if (!strcasecmp(*p,"FMDOWN+"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.dl_f_month+=i; else se = 1;
	 } else if (!strcasecmp(*p,"FYUP+"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.ul_f_year+=i; else se = 1;
	 } else if (!strcasecmp(*p,"FYUP-"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.ul_f_year-=i; else se = 1;
	 } else if (!strcasecmp(*p,"FYUP"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.ul_f_year=i; else se = 1;
	 } else if (!strcasecmp(*p,"FYDOWN-"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.dl_f_year-=i; else se = 1;
	 } else if (!strcasecmp(*p,"FYDOWN"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.dl_f_year=i; else se = 1;
	 } else if (!strcasecmp(*p,"FYDOWN+"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.dl_f_year+=i; else se = 1;
	 } else if (!strcasecmp(*p,"FDUP+"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.ul_f_day+=i; else se = 1;
	 } else if (!strcasecmp(*p,"FDUP-"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.ul_f_day-=i; else se = 1;
	 } else if (!strcasecmp(*p,"FDUP"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.ul_f_day=i; else se = 1;
	 } else if (!strcasecmp(*p,"FDDOWN-"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.dl_f_day-=i; else se = 1;
	 } else if (!strcasecmp(*p,"FDDOWN"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.dl_f_day=i; else se = 1;
	 } else if (!strcasecmp(*p,"FDDOWN+"))
	 {   if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
            if (sscanf(q,SI32,&i)) usr.dl_f_day+=i; else se = 1;
	 } 

	 /* SPEEDS */
	 else if(!strcasecmp(*p, "KBSUP")) {  
	    if(!allowed(c->user, R_USER_EDIT_SPEEDS, u)) pd = 1; 
	    else if (sscanf(q, SI64, &l)) {
	       usr.ul_tm_total = l ? ((usr.ul_total / 1024) / l) : 0; 
	    }
	    else se = 1;
	 } 
	 else if(!strcasecmp(*p, "KBSDOWN")) {  
	    if(!allowed(c->user, R_USER_EDIT_SPEEDS, u)) pd = 1; 
	    else if (sscanf(q, SI64, &l)) {
	       usr.dl_tm_total = l ? ((usr.dl_total / 1024) / l) : 0; 
	    }
	    else se = 1;
	 } 
	 else if(!strcasecmp(*p, "KBSWKDOWN")) {  
	    if(!allowed(c->user, R_USER_EDIT_SPEEDS, u)) pd = 1; 
	    else if (sscanf(q, SI64, &l)) {
	       usr.dl_tm_week = l ? ((usr.dl_week / 1024) / l) : 0; 
	    }
	    else se = 1;
	 } 
	 else if(!strcasecmp(*p, "KBSWKUP")) {  
	    if(!allowed(c->user, R_USER_EDIT_SPEEDS, u)) pd = 1; 
	    else if (sscanf(q, SI64, &l)) {
	       usr.ul_tm_week = l ? ((usr.ul_week / 1024) / l) : 0; 
	    }   
	    else se = 1;   
	 } 
	 else if(!strcasecmp(*p, "KBSMDOWN")) {  
	    if(!allowed(c->user, R_USER_EDIT_SPEEDS, u)) pd = 1; 
	    else if (sscanf(q, SI64, &l)) {
	       usr.dl_tm_month =  l ? ((usr.dl_month / 1024) / l) : 0; 
	    }   
	    else se = 1;
	 } 
	 else if(!strcasecmp(*p, "KBSMUP")) {  
	    if(!allowed(c->user, R_USER_EDIT_SPEEDS, u)) pd = 1; 
	    else if (sscanf(q, SI64, &l)) {
	       usr.ul_tm_month = l ? ((usr.ul_month / 1024) / l) : 0; 
	    }   
	    else se = 1;
	 } 
	 else if(!strcasecmp(*p, "KBSYDOWN")) {  
	    if(!allowed(c->user, R_USER_EDIT_SPEEDS, u)) pd = 1; 
	    else if (sscanf(q, SI64, &l)) {
	       usr.dl_tm_year = l ? ((usr.dl_year / 1024) / l) : 0; 
	    }
	    else se = 1;
	 } 
	 else if(!strcasecmp(*p, "KBSYUP")) {  
	    if(!allowed(c->user, R_USER_EDIT_SPEEDS, u)) pd = 1; 
	    else if (sscanf(q, SI64, &l)) {
	       usr.ul_tm_year = l ? ((usr.ul_year / 1024) / l) : 0; 
	    }
	    else se = 1;
	 } 
	 else if(!strcasecmp(*p, "KBSDDOWN")) {  
	    if(!allowed(c->user, R_USER_EDIT_SPEEDS, u)) pd = 1; 
	    else if (sscanf(q, SI64, &l)) {
	       usr.dl_tm_day =  l ? ((usr.dl_day / 1024) / l) : 0; 
	    }
	    else se = 1;
	 } 
	 else if(!strcasecmp(*p, "KBSDUP")) {  
	    if(!allowed(c->user, R_USER_EDIT_SPEEDS, u)) pd = 1; 
	    else if (sscanf(q, SI64, &l)) {
	       usr.ul_tm_day = l ? (usr.ul_day / 1024) / l : 0; 
	    }
	    else se = 1;

         /* NUKES */
	 } else if(!strcasecmp(*p,"FNUKED"))
	 {  if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
           if (sscanf(q,SI32,&i)) usr.f_nuked_total=i; else se = 1;
	 } else if(!strcasecmp(*p,"FWKNUKED"))
	 {  if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
           if (sscanf(q,SI32,&i)) usr.f_nuked_week=i; else se = 1;
	 } else if(!strcasecmp(*p,"FMNUKED"))
	 {  if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
           if (sscanf(q,SI32,&i)) usr.f_nuked_month=i; else se = 1;
	 } else if(!strcasecmp(*p,"FYNUKED"))
	 {  if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
           if (sscanf(q,SI32,&i)) usr.f_nuked_year=i; else se = 1;
	 } else if(!strcasecmp(*p,"FDNUKED"))
	 {  if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
           if (sscanf(q,SI32,&i)) usr.f_nuked_day=i; else se = 1;
	 } else if(!strcasecmp(*p,"TNUKED"))
	 {  if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
           if (sscanf(q,SI32,&i)) usr.t_nuked_total=i; else se = 1;
	 } else if(!strcasecmp(*p,"TWKNUKED"))
	 {  if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
           if (sscanf(q,SI32,&i)) usr.t_nuked_week=i; else se = 1;
	 } else if(!strcasecmp(*p,"TMNUKED"))
	 {  if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
           if (sscanf(q,SI32,&i)) usr.t_nuked_month=i; else se = 1;
	 } else if(!strcasecmp(*p,"TYNUKED"))
	 {  if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
           if (sscanf(q,SI32,&i)) usr.t_nuked_year=i; else se = 1;
	 } else if(!strcasecmp(*p,"TDNUKED"))
	 {  if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
           if (sscanf(q,SI32,&i)) usr.t_nuked_day=i; else se = 1;
	 } else if(!strcasecmp(*p,"CNUKED"))
	 {  if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
           if (sscanf(q,SI64,&l)) usr.c_nuked_total=l * 0x100000; else se = 1;
	 } else if(!strcasecmp(*p,"CWKNUKED"))
	 {  if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
           if (sscanf(q,SI64,&l)) usr.c_nuked_week=l * 0x100000; else se = 1;
	 } else if(!strcasecmp(*p,"CMNUKED"))
	 {  if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
           if (sscanf(q,SI64,&l)) usr.c_nuked_month=l * 0x100000; else se = 1;
	 } else if(!strcasecmp(*p,"CYNUKED"))
	 {  if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
           if (sscanf(q,SI64,&l)) usr.c_nuked_year=l * 0x100000; else se = 1;
	 } else if(!strcasecmp(*p,"CDNUKED"))
	 {  if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
           if (sscanf(q,SI64,&l)) usr.c_nuked_day=l * 0x100000; else se = 1;
	 } else if(!strcasecmp(*p,"MNUKED"))
	 {  if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
           if (sscanf(q,SI64,&l)) usr.b_nuked_total=l * 0x100000; else se = 1;
	 } else if(!strcasecmp(*p,"MWKNUKED"))
	 {  if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
           if (sscanf(q,SI64,&l)) usr.b_nuked_week=l * 0x100000; else se = 1;
	 } else if(!strcasecmp(*p,"MMNUKED"))
	 {  if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
           if (sscanf(q,SI64,&l)) usr.b_nuked_month=l * 0x100000; else se = 1;
	 } else if(!strcasecmp(*p,"MYNUKED"))
	 {  if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
           if (sscanf(q,SI64,&l)) usr.b_nuked_year=l * 0x100000; else se = 1;
	 } else if(!strcasecmp(*p,"MDNUKED"))
	 {  if(!allowed(c->user,R_USER_EDIT_STATS,u)) pd = 1; else
           if (sscanf(q,SI64,&l)) usr.b_nuked_day=l * 0x100000; else se = 1;
         /* end nuke section */
	 } 
	 
	 /* LOGINS */
	 else if (!strcasecmp(*p, "LOGINS")) { 
	    if (!allowed(c->user, R_USER_EDIT_LOGINS, u)) pd = 1; 
	    else if (sscanf(q, SI32, &i) && (i >= 0) && (i <= 255)) {
               usr.class = (usr.class & ~UL_MASK) | (i << 8); 
	    }   
	    else se = 1;
	 } 

         /* CLASS */
	 else if (!strcasecmp(*p, "CLASS") && (strlen(q) == 1)) { 
	    usr.class &= ~UC_MASK;
/* don't check slots on class change, can't remember why this was added
	    if (tolower(*q) != 't' && !templateuser(&usr)) {
	       if ((usr.gid != -1) && (has_group_free_slots(usr.gid, str, sizeof(str)) == 1)) { 
                  server_reply(c, "- Group is full."); 
                  return 0; 
	       }
            }
*/
	    switch (tolower(*q)) { 
	       case 'a' :
        	  if (!allowed(c->user,R_USER_ADD_ANON,NULL)) pd = 1;
        	  else usr.class |= UC_ANON;
        	  break;
               case 'g' :
        	  if (!allowed(c->user,R_USER_ADD_GO,NULL)) pd = 1;
        	  else usr.class |= UC_GROUPOP;
        	  break;
               case 'n' :
        	  if (!allowed(c->user,R_USER_ADD,NULL)) pd = 1;
        	  else usr.class |= UC_NORMAL;
        	  break;
               case 'o' :
        	  if (!allowed(c->user,R_USER_ADD_SO,NULL)) pd = 1;
        	  else usr.class |= UC_SITEOP;
        	  break;
               case 'r' :
               case 's' :
        	  if (!superuser(c->user)) pd = 1;
        	  else usr.class |= UC_ROOT;
        	  break;
               case 't' :
        	  if (!allowed(c->user,R_USER_ADD_TMP,NULL)) pd = 1;
        	  else usr.class |= UC_TEMPLATE;
        	  break;
               default  :
        	  se = 1;
            }
	 } 
	 
         /* FLAGS+ */	 
	 else if (!strcasecmp(*p, "FLAGS+") && (strlen(q) == 1)) { 
	    if (!allowed(c->user,R_USER_EDIT,u)) pd = 1; 
	    else switch (tolower(*q)) { 
	       case 'c' : usr.class |= UF_COLOR; break;
               case 'd' : usr.class |= UF_DISABLED; break;
               case 'i' : usr.class |= UF_INVISIBLE; break;
               case 'p' : usr.class |= UF_NOPASSCH; break;
               case '-' : break;
               default  : se = 1;
            }
	 } 
	 
	 /* FLAGS- */
	 else if (!strcasecmp(*p, "FLAGS-") && (strlen(q) == 1)) { 
	    if (!allowed(c->user,R_USER_EDIT,u)) pd = 1; 
	    else switch (tolower(*q)) { 
	       case 'c': usr.class &= ~UF_COLOR; break;
               case 'd': if(has_group_free_slots(usr.gid,str,sizeof(str)) == 1) {
                	   server_reply(c,str);
                	   return 0;
                	 } 
			 else usr.class &= ~UF_DISABLED; 
			 break;
               case 'i': if(has_group_free_slots(usr.gid, str, sizeof(str)) == 1) {
                	   server_reply(c,str);
                	   return 0;
                	 } 
			 else usr.class &= ~UF_INVISIBLE; 
			 break;
               case 'p': usr.class &= ~UF_NOPASSCH; break;
               case '-': break;
               default : se = 1;
            }
	 } 
	 
	 /* UL_MIN */
	 else if (!strcasecmp(p[0], "UL_MIN")) { 
	    if (!allowed(c->user, R_USER_EDIT_SPEEDS, u)) pd = 1; 
	    else if (!sscanf(q, SI32, &j)) se = 1;
            else usr.ul_speed = (usr.ul_speed & 0xFFFF0000) | j;
	 } 
	 
	 /* UL_MAX */
	 else if (!strcasecmp(p[0], "UL_MAX")) { 
	    if (!allowed(c->user, R_USER_EDIT_SPEEDS, u)) pd = 1; 
	    else if (!sscanf(q,SI32,&j)) se = 1;
            else usr.ul_speed = (usr.ul_speed & 0xFFFF) | (j << 16);
	 } 
	 
	 /* DL_MIN */
	 else if (!strcasecmp(p[0], "DL_MIN")) { 
	    if (!allowed(c->user, R_USER_EDIT_SPEEDS, u)) pd = 1; 
	    else if (!sscanf(q, SI32, &j)) se = 1;
            else usr.dl_speed = (usr.dl_speed & 0xFFFF0000) | j;
	 } 
	 
	 /* DL_MAX */
	 else if (!strcasecmp(p[0], "DL_MAX")) { 
	    if (!allowed(c->user,R_USER_EDIT_SPEEDS,u)) pd = 1; 
	    else if (!sscanf(q,SI32,&j)) se = 1;
            else usr.dl_speed = (usr.dl_speed & 0xFFFF) | (j << 16);
	 } 
	 
	 /* PASSWD */
	 else if (!strcasecmp(*p, "PASSWD")) { 
	    if (!allowed(c->user,R_USER_PASSWD,u) || 
  	       (strlen(q) < sizeof(usr.password) + 1)) {
               if (strlen(q) == 0) usr.password[0] = '\0'; 
	       else { 
	          salt[0] = salt_a[time(NULL) & 0x3F];
        	  salt[1] = salt_a[(time(NULL) >> 6) & 0x3F];
        	  strlcpy(usr.password, (char *)crypt(q, salt), sizeof(usr.password));
               } 
	    }
	    else pd = 1;
	 } 
	 else se = 1;
      } else se = 1;
      if (pd || se) n = 1;
   }
   if (pd) { 
      server_reply(c, "-Permission denied.");
      return 0;
   } 
   else if (se) { 
      server_reply(c, "-Syntax error.");
      return 0;
   } 
   else if (rf) { 
      server_reply(c, "-IP-range full.");
      return 0;
   }

   /* if uid is to change, re-sort users */
   if (usr.uid != u->uid) { 
      user_count--;
      if (u-user < user_count) memmove(u,u+1,(user_count-(u-user))*sizeof(*u));
      for (i = 0, ch = child; i < children; i++, ch++) {
	if (ch->user == u) ch->user = (struct user_s*)-1;
	else if (ch->user > u) ch->user--;
      }
      u = insert_user(&usr);
      for (i = 0, ch = child; i < children; i++, ch++) {
	 if (ch->user == (struct user_s*)-1) ch->user = u;
      }
   } 
   else { 
      *u = usr;
      user_dirty++;
   }
   server_log_local(c, LG_SECURITY, "USER %s was modified by %s", u->name, 
      c->user->name);
   server_reply(c, "+User changed.");
   return 0;
}


/* print user */
void print_user(struct user_s* u, char* str)
{ 
   int i;

   sprintf(str,
      "+%s\t"SI32"\t"SI32"\t"SI32"\t"SI32"\t"SI32"\t"
      SI64"\t"SI32"\t"SI64"\t"SITIME"\t"SITIME"\t%s\t"
      SI32"\t"SI64"\t"SI64"\t"
      SI32"\t"SI64"\t"SI64"\t"
      SI32"\t"SI64"\t"SI64"\t"
      SI32"\t"SI64"\t"SI64"\t"
      SI32"\t"SI64"\t"SI64"\t"
      SI32"\t"SI64"\t"SI64"\t"
      SI32"\t"SI64"\t"SI64"\t"
      SI32"\t"SI64"\t"SI64"\t"
      SI32"\t"SI64"\t"SI64"\t"
      SI32"\t"SI64"\t"SI64"\t"
      SI32"\t"SI64"\t"SI64"\t"SI32"\t"
      SI32"\t"SI64"\t"SI64"\t"SI32"\t"
      SI32"\t"SI64"\t"SI64"\t"SI32"\t"
      SI32"\t"SI64"\t"SI64"\t"SI32"\t"
      SI32"\t"SI64"\t"SI64"\t"SI32"\t"
      SI64,

      u->name, u->uid, u->gid, u->limit, u->ul_speed, u->dl_speed,
      u->credit, u->class, u->rights, u->last_login, u->creation_time,
      u->creator && strlen(u->creator) > 0 ? u->creator : "unknown",

      u->dl_f_day,   u->dl_day,   u->dl_tm_day,
      u->ul_f_day,   u->ul_day,   u->ul_tm_day,
      u->dl_f_week,  u->dl_week,  u->dl_tm_week,
      u->ul_f_week,  u->ul_week,  u->ul_tm_week,
      u->dl_f_month, u->dl_month, u->dl_tm_month,
      u->ul_f_month, u->ul_month, u->ul_tm_month,
      u->dl_f_year,  u->dl_year,  u->dl_tm_year,
      u->ul_f_year,  u->ul_year,  u->ul_tm_year,
      u->dl_f_total, u->dl_total, u->dl_tm_total,
      u->ul_f_total, u->ul_total, u->ul_tm_total,

      u->f_nuked_day,   u->b_nuked_day,   u->c_nuked_day,   u->t_nuked_day,
      u->f_nuked_week,  u->b_nuked_week,  u->c_nuked_week,  u->t_nuked_week,
      u->f_nuked_month, u->b_nuked_month, u->c_nuked_month, u->t_nuked_month,
      u->f_nuked_year,  u->b_nuked_year,  u->c_nuked_year,  u->t_nuked_year,
      u->f_nuked_total, u->b_nuked_total, u->c_nuked_total, u->t_nuked_total,
      u->wkly_alot
   );

   str += strlen(str);
   for (i = 0; i < 8; i++) { 
      sprintf(str, "\t"SI32, u->gids[i]);
      str += strlen(str);
   }
   for (i = 0; i < 10; i++) { 
      *(str++) = '\t';
      print_ident(str, u->idents[i]);
      str += strlen(str);
   }
   str += strlen(str);
   sprintf(str, "\t\"%s\"", u->info[0] ? u->info : "-");
}


/* user get */
int f4adp_usr_get(struct child_s* c, char** p, int n)
{ 
   char           str[1024];
   struct user_s* u;

   u = get_user_by_name(p[0]);
   if (!u) { 
      server_reply(c, "-User does not exist.");
      return 0;
   }
   if (!allowed(c->user, R_USER_READ, u)) { 
      server_reply(c,"-Permission denied.");
      return 0;
   }
   print_user(u, str);
   server_reply(c, str);
   return 0;
}


/* user kick (kill all sessions) */
int usr_kick(struct child_s** c, struct user_s* u, int signo)
{ struct child_s* ch;
  int             found;
  int             i;

  for (found=0, i=0, ch=child; i<children; ) if (ch->user == u)
  { found++;
    if (ch->pid != -1) kill(ch->pid,signo);
    if (*c)
    { if (ch == *c) *c = NULL; else if (ch < *c) (*c)--;
    }
    child_logout(ch,CL_KICK);
  } else
  { ch++;
    i++;
  }
  return found;
}


/* user kick (kill all sessions) */
int f4adp_usr_kick(struct child_s* c, char** p, int n)
{ struct child_s* ch;
  int             signo;
  struct user_s*  u;

  if (n == 2)
  { if (!sscanf(p[1],"%i",&signo)) signo = -1;
  } else signo = SIGQUIT;
  if (signo == -1)
  { server_reply(c,"-Syntax error");
    return 0;
  }
  u = get_user_by_name(p[0]);
  if (!u)
  { server_reply(c,"-User does not exist");
    return 0;
  }
  if (!allowed(c->user,R_USER_KICK,u))
  { server_reply(c,"-Permission denied");
    return 0;
  }
  ch = c;
  n = usr_kick(&ch,u,signo);
  if (ch)
  { server_reply(c,n ? "+User kicked" : "-User is not logged in");
    return c - ch;
  }
  return -1;
}

/* user kill session */
int f4adp_usr_kill(struct child_s* c, char** p, int n)
{ 
   struct child_s* ch;
   int             i;
   pid_t           pid;
   int             signo;

   if (n == 2) { 
      if (!sscanf(p[1],"%i",&signo)) signo = -1;
   } 
   else signo = SIGQUIT;
   if ((signo == -1) || !sscanf(p[0],"%ld",(long *)&pid)) { 
      server_reply(c,"-Syntax error");
      return 0;
   }
   for (i=0, ch=child; i<children; i++, ch++) {
      if (ch->pid == pid) { 
         if (!allowed(c->user, R_USER_KICK, ch->user)) { 
	    server_reply(c,"-Permission denied");
	    return 0;
	 }
	 if (pid != -1) kill(pid,signo);
	 server_reply(c,"+Session killed");
	 child_logout(ch,CL_KICK);
	 return c == ch ? -1 : (c < ch ? 0 : 1);
      }
   }
   server_reply(c,"-Session does not exist");
   return 0;
}


/* user list */
int f4adp_usr_ls(struct child_s* c, char** p, int n)
{ 
   /* this is a really lousy fix for the problem with too
    * much users, dynamic memory would be better here but
    * this would result in rewrite of bigger parts of the
    * source - primemover 
    */
   char           str[1024 * 10];
   struct user_s* u;

   for (n = 0, u = user; n < user_count; n++, u++) {
      if (allowed(c->user, R_USER_READ, u)) { 
         print_user(u, str);
	 server_reply(c, str);
      }
   }
   server_reply(c,"+");
   return 0;
}


/* user list restricted */
int f4adp_usr_lsr(struct child_s* c, char** p, int n)
{ 
   char           str[32];
   struct user_s* u;   

   g_snprintf(str, sizeof(str), "%i", user_count);
   server_reply(c, str);
   if (!user_count) return 0;
   if (!fgets_to(str, sizeof(str), c->rdp, 5)) return 0;
   if (*str != '+') return 0;
   for (n = 0, u = user; n<user_count; n++) { 
      g_snprintf(str, sizeof(str), "%s "SI32, u->name, u->uid);
      server_reply(c,str);
      u++;
   }
   return 0;
}
/* send user tagline */
int f4adp_usr_nfo(struct child_s* c, char** p, int n)
{
    char str[35];
    
    snprintf(str, sizeof(str), "+%s", *c->user->info ? c->user->info : "-");
    server_reply(c, str);
    return 0;
}

/* user password */
int f4adp_usr_pwd(struct child_s* c, char** p, int n)
{ 
   char           salt[2];
   struct user_s* u;

   if (strcasecmp(p[0], c->user->name)) { 
      u = get_user_by_name(p[0]);
      if (!u) { 
         server_reply(c,"-User does not exist");
         return 0;
      }
      if (!allowed(c->user,R_USER_PASSWD, u)) { 
         server_reply(c,"-Permission denied");
         return 0;
      }
   } 
   else { 
      u = c->user;
      if (anonuser(u) || no_pass_change(u)) { 
         server_reply(c, "-Permission denied");
         return 0;
      }
   }
   if ((n == 1) || !p[1][0]) { 
      /* reset password, if no password given or password is null string */
      u->password[0] = '\0';
      server_log_local(c, LG_SECURITY, "PASSWORD of %s was cleared by %s", 
         u->name, u == c->user ? c->user->name : "himself");
      server_reply(c,"+Password cleared");
   } 
   else { 
      /* store crypted password */
      salt[0] = salt_a[time(NULL) & 0x3F];
      salt[1] = salt_a[(time(NULL) >> 6) & 0x3F];
      strlcpy(u->password, (char *)crypt(p[1], salt), sizeof(u->password));
      server_log_local(c, LG_SECURITY, "PASSWORD of %s was changed by %s", 
         u->name, u == c->user ? c->user->name : "himself");
      server_reply(c, "+Password changed");
   }
   user_dirty++;
   return 0;
}


/* user remove */
int f4adp_usr_rm(struct child_s* c, char** p, int n)
{ 
   struct child_s* ch = child;
   int             i;
   struct user_s*  u;
   struct sstat_s  *st;

   u = get_user_by_name(p[0]);
   if (!u) { 
      server_reply(c,"-User does not exist");
      return 0;
   }
   if (!allowed(c->user, R_USER_DELETE, u)) { 
      server_reply(c,"-Permission denied");
      return 0;
   }
   for (i = 0; i < children; i++, ch++) {
      if (ch->user) { 
         if (ch->user == u) { 
	    if (ch->pid != -1) kill(ch->pid, SIGKILL);
	    child_logout(ch, CL_KICK);
	 } 
	 else if (ch->user > u) ch->user--;
      }
   }
   /* delete section stats for user */
   for (i = 0; i < section_count; i++) {
      st = get_stat_by_uid(section+i, u->uid);
      if (st) {
         memmove(st, st+1, (--((section+i)->stat_count)-(st-
	    ((section+i)->stat)))*sizeof(struct sstat_s));
         section_stat_dirty++;     
      }  
   }
   memmove(u,u+1,(--user_count-(u-user))*sizeof(struct user_s));
   user_dirty++;
   server_log_local(c, LG_SECURITY, "USER %s was removed by %s", 
      u->name, c->user->name);
   server_reply(c,"+User removed");
   return 0;
}


/* user reset */
int f4adp_usr_rst(struct child_s* c, char** p, int n)
{ 
   int            r_time = 0;
   int            r_what = 0;
   int            se     = 0;
   struct user_s* u;

   if (!strcasecmp(p[0],"all")) r_what = 0x0F; 
   else if (!strcasecmp(p[0],"up")) r_what = 0x01; 
   else if (!strcasecmp(p[0],"down")) r_what = 0x02; 
   else if (!strcasecmp(p[0],"nukes")) r_what = 0x04; 
   else if (!strcasecmp(p[0],"credit")) r_what = 0x08; 
   else se = 1;
   if (n == 2) { 
      if (r_what & 0x07) { 
	 if (!strcasecmp(p[1],"all")) r_time = 0x01f;
	 else if (!strcasecmp(p[1],"week")) r_time = 0x01; 
	 else if (!strcasecmp(p[1],"month")) r_time = 0x02; 
	 else if (!strcasecmp(p[1],"year")) r_time = 0x04; 
	 else if (!strcasecmp(p[1],"total")) r_time = 0x08; 
	 else if (!strcasecmp(p[1],"day")) r_time = 0x10; 
	 else se = 1;
	 if (r_time != 0x01f) r_what &= 0x07;
      } 
      else se = 1;
   }
   if (se) { 
      server_reply(c,"-Syntax error");
      return 0;
   }
   for (n = 0, u = user; n < user_count; n++, u++) {
      if (!templateuser(u)) { 
         if (r_what & 0x01) { 
	    if (r_time & 0x01) u->ul_f_week = u->ul_week = u->ul_tm_week = 0;
	    if (r_time & 0x02) u->ul_f_month = u->ul_month = u->ul_tm_month = 0;
	    if (r_time & 0x04) u->ul_f_year = u->ul_year = u->ul_tm_year = 0;
	    if (r_time & 0x08) u->ul_f_total = u->ul_total = u->ul_tm_total = 0;
	    if (r_time & 0x10) u->ul_f_day = u->ul_day = u->ul_tm_day = 0;
	 }
	 if (r_what & 0x02) { 
	    if (r_time & 0x01) u->dl_f_week = u->dl_week = u->dl_tm_week = 0;
	    if (r_time & 0x02) u->dl_f_month = u->dl_month = u->dl_tm_month = 0;
	    if (r_time & 0x04) u->dl_f_year = u->dl_year = u->dl_tm_year = 0;
	    if (r_time & 0x08) u->dl_f_total = u->dl_total = u->dl_tm_total = 0;
	    if (r_time & 0x10) u->dl_f_day = u->dl_day = u->dl_tm_day = 0;
	 }
	 if (r_what & 0x04) { 
	    if (r_time & 0x01) u->c_nuked_week = u->b_nuked_week = 
	       u->f_nuked_week = u->t_nuked_week = 0;
	    if (r_time & 0x02) u->c_nuked_month = u->b_nuked_month = 
	       u->f_nuked_month = u->t_nuked_month = 0;
	    if (r_time & 0x04) u->c_nuked_year = u->b_nuked_year =
	       u->f_nuked_year = u->t_nuked_year = 0;
	    if (r_time & 0x08) u->c_nuked_total = u->b_nuked_total = 
	       u->f_nuked_total = u->t_nuked_total = 0;
	    if (r_time & 0x10) u->c_nuked_day = u->b_nuked_day =
	       u->f_nuked_day = u->t_nuked_day = 0;
	}
	if (r_what & 0x08) u->credit = 0;
      }
   }
   user_dirty++;
   server_reply(c,"+Reset successful");
   return 0;
}


/* user set  XXX need to be changed to use the gui */
int f4adp_usr_set(struct child_s* c, char** p, int n)
{ 
   struct child_s* ch;
   int             i;
   gint64           l;
   int             pd;
   struct user_s*  u;
   struct user_s   usr;

   u = get_user_by_name(p[0]);
   if (!u) { 
      server_reply(c,"-User does not exist");
      return 0;
   }
   pd = !allowed(c->user,R_USER_EDIT,u);
   if (!pd) { 
      if (strcasecmp(p[0], p[1])) {
	 if (get_user_by_name(p[1])) { 
	    server_reply(c,"-User already exists");
	    return 0;
	 }
      }
      usr = *u;
      strlcpy(usr.name, p[1], sizeof(usr.name));
      sscanf(p[2],SI32,&usr.uid);
      if (usr.uid != u->uid) if (!allowed(c->user,R_USER_EDIT_UID,NULL)) pd = 1;
      sscanf(p[3],SI32,&usr.gid);
      if (groupop(c->user) && !groupopof(c->user,usr.gid)) pd = 1;
      sscanf(p[4],SI32,&usr.limit);
      if (usr.limit != u->limit) if (!allowed(c->user,R_USER_EDIT_CREDIT,NULL)) pd = 1;
      if (sscanf(p[5],SI64,&l))
      { if (l) if (!allowed(c->user,R_USER_EDIT_CREDIT,NULL)) pd = 1;
	usr.credit += l;
      } else
      { if (l) if (!allowed(c->user,R_USER_EDIT_CREDIT,NULL)) pd = 1;
	usr.credit = 0;
      }
      sscanf(p[6],SI32,&usr.class);
      if ((usr.class & UC_MASK) != (u->class & UC_MASK))
	switch (usr.class & UC_MASK)
      { case UC_NORMAL   :
          if (!allowed(c->user,R_USER_ADD,NULL)) pd = 1; break;
	case UC_GROUPOP  :
          if (!allowed(c->user,R_USER_ADD_GO,NULL)) pd = 1; break;
	case UC_SITEOP   :
          if (!allowed(c->user,R_USER_ADD_SO,NULL)) pd = 1; break;
	case UC_TEMPLATE :
          if (!allowed(c->user,R_USER_ADD_TMP,NULL)) pd = 1; break;
	case UC_ANON :
          if (!allowed(c->user,R_USER_ADD_ANON,NULL)) pd = 1; break;
	default          :
          if (!superuser(c->user)) pd = 1;
      }
      if (usr.class >> 8 != u->class >> 8)
	if (!allowed(c->user,R_USER_EDIT_LOGINS,NULL)) pd = 1;
      sscanf(p[7],SI64,&usr.rights);
      if (usr.rights != u->rights) if (!superuser(c->user))
      { if (c->user == u) pd = 1; else
	if (!allowed(c->user,R_USER_EDIT_RIGHTS,NULL)) pd = 1; else
	/* you must have the rights that you want to grant/revoke */
	if ((c->user->rights | (usr.rights ^ u->rights)) != c->user->rights)
          pd = 1;
      }
      if (strlen(p[8]) >= sizeof(usr.info))
      { server_reply(c,"-User information is too long.");
	return 0;
      }
      strlcpy(usr.info, strcmp(p[8],"-") ? p[8] : "", sizeof(usr.info));
      if (strcmp(usr.info,u->info))
	if (!allowed(c->user,R_USER_EDIT_INFO,NULL)) pd = 1;
      sscanf(p[9],SI32,&usr.ul_speed);
      if (usr.ul_speed != u->ul_speed) if (!allowed(c->user,R_USER_EDIT_SPEEDS,NULL)) pd = 1;
      sscanf(p[10],SI32,&usr.dl_speed);
      if (usr.dl_speed != u->dl_speed) if (!allowed(c->user,R_USER_EDIT_SPEEDS,NULL)) pd = 1;
      for (n=0; n<8; n++)
      { sscanf(p[n+11],SI32,&usr.gids[n]);
	if (usr.gids[n] != u->gids[n])
          if (!allowed(c->user,R_USER_EDIT_GRPS,NULL)) pd = 1;
      }
      for (n=0; n<10; n++)
      { get_iprange(&p[n+19],usr.idents[n]);
	if (memcmp(usr.idents[n],u->idents[n],sizeof(usr.idents[0])))
          if (!allowed(c->user,R_USER_EDIT_IP,NULL)) pd = 1;
      }
   }
   if (pd) { 
      server_reply(c,"-Permission denied");
      return 0;
   }
   /* if uid is to change, re-sort users */
   if (usr.uid != u->uid)
   { user_count--;
     if (u-user < user_count) memmove(u,u+1,(user_count-(u-user))*sizeof(*u));
     for (i=0, ch=child; i<children; i++, ch++)
       if (ch->user == u) ch->user = (struct user_s*)-1;
       else if (ch->user > u) ch->user--;
     u = insert_user(&usr);
     for (i=0, ch=child; i<children; i++, ch++)
       if (ch->user == (struct user_s*)-1) ch->user = u;
   } else
   { *u = usr;
     user_dirty++;
   }
   server_reply(c,"+User changed");
   return 0;
}
int f4adp_usr_credits(struct child_s* c, char** p, int n)
{  
   char str[56];

   g_snprintf(str, sizeof(str), "+" SI64, c->user->credit);
   server_reply(c, str);
   return 0; 
}

/* EOF */
