/* sections.c
 *
 * This file is part of OpenFTPD (based on FTP4ALL)
 *
 * This program is free software and was developed under the terms of
 * GNU General Public License. Reas LICENSE for more information.
 *
 * Content of this module: most section related functions
 *
 */

#include <sys/types.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include "externsd.h"
#include "log.h"
#include "serverd.h"
#include "utils.h"    
#include "sections.h"
#include "user.h"
#include "group.h"

#define D_BOOL  1
#define D_INT   2
#define D_CHAR  3
#define D_HEX   4 /* just only for output format */

#define YES     1
#define NO      0


/*
 * key table must be sorted alphabetically by key for binary search
 * all keys ftpd have to read from section config files should go here 
 */

struct key_type_s {
   char key[32];
   int  value;
   char def[50];
} key_type[] = {
   { "allow_check"                      , D_CHAR , "\0"           },
   { "allow_dupe"                       , D_CHAR , "\0"           },
   { "announce_badfile"                 , D_BOOL , "no"           },
   { "announce_complete"                , D_BOOL , "no"           },
   { "announce_deldir"                  , D_BOOL , "no"           },
   { "announce_delfile"                 , D_BOOL , "no"           },
   { "announce_http"                    , D_BOOL , "no"           },
   { "announce_imdb"                    , D_BOOL , "no"           },
   { "announce_joinrace"                , D_BOOL , "no"           },
   { "announce_newdir"                  , D_BOOL , "no"           },
   { "announce_sample"                  , D_BOOL , "no"           },
   { "announce_update"                  , D_BOOL , "no"           },
   { "calc_checksum"                    , D_BOOL , "no"           }, 
   { "check_sfv"                        , D_BOOL , "no"           },
   { "check_zip"                        , D_BOOL , "no"           },
   { "complete_dir"                     , D_CHAR , "[ABC]-COMPLETE-[ABC]"	},
   { "copyinfofile"                     , D_BOOL , "no"           }, 
   { "dccinfo"                          , D_BOOL , "yes"          },
   { "dccpath"                          , D_CHAR , "\0"           },
   { "del_empty_nuke"                   , D_BOOL , "no"           }, 
   { "deny_dirs"                        , D_CHAR , "\0"           },
   { "different_credit"                 , D_BOOL , "no"           },
   { "dir_dupecheck"                    , D_BOOL , "no"           },
   { "download_limit"                   , D_INT  , "-1"           },
   { "dupecheck"                        , D_BOOL , "yes"          },
   { "dupe_filetypes"                   , D_CHAR , "\0"           },
   { "empty_nuke"                       , D_INT  , "25"           },
   { "enforce_sfv"                      , D_BOOL , "no"           },
   { "internal_dupecheck"               , D_BOOL , "no"           },
   { "lock_complete"                    , D_BOOL , "no"           }, 
   { "no_global_stats"                  , D_BOOL , "no"           },
   { "nukeprefix"                       , D_CHAR , "_NUKED_"      },
   { "prepath"                          , D_CHAR , "\0"           },
   { "resolvesymlinks"                  , D_BOOL , "yes"          },
   { "show_dirstat"                     , D_BOOL , "yes"          },
   { "show_newdirs"                     , D_BOOL , "yes"          },
   { "siteinfofile"                     , D_CHAR , "\0"           },
   { "status_dir"                       , D_BOOL , "yes"          },
   { "statusfilepath"                   , D_CHAR , "\0"           },
   { "umask_anon"                       , D_HEX  , "0x230000"     },
   { "umask_root"                       , D_HEX  , "0x7F1300"     },
   { "umask_user"                       , D_HEX  , "0x7F1300"     },
   { "undupe_on_delete"                 , D_BOOL , "yes"          },
   { "upload_limit"                     , D_INT  , "-1"           },
   { "zipcommentfile"                   , D_CHAR , "\0"           }
}; 


/* load and handle etc/sections.cfg */
int load_global_section_config(void)
{ 
   FILE             *file;
   int              i;
   int              found_stat;
   int              found_config;
   char             str[BUFFER_LEN];   
   struct section_s *s;
 
   /* First we get default one for root path */
   alloc_section();
   if (section_count >= section_size) { 
      errlog("out of memory (sections 0)");
      return -1;
   }

   strlcpy(section->section, "default", sizeof(section->section));
   strlcpy(section->config_file, "default.cfg", sizeof(section->config_file));
   strlcpy(section->stat_file, "default.stat", sizeof(section->stat_file));
   strlcpy(section->path, "/", sizeof(section->path));
   section_count++;
   section->config_size = section->config_count = 0;
   section->stat_size = section->stat_count = 0;
   section->stat = NULL;
   section->config = NULL;
   
   if (load_config_for_single_section(section) == -1) { 
      errlog("Error reading default section");
      return -1;
   }

   /* load default values for keys not defined in default.cfg */
   for (i = 0; i < (sizeof(key_type)/sizeof(struct key_type_s)); i++) {
      if (search_key(key_type[i].key, section, NULL)) {
         g_snprintf(str, sizeof(str), "%s %s\n", key_type[i].key, key_type[i].def);
	 parse_config_for_section_line(str, section);
      }
   }

   if (load_stat_file(section) == -1) { 
       if (init_stat_file(section) == -1) {
          errlog("Error creating default stat file");
	  return -1;
       }
       save_stat_file(section);
   }

   /* Now, read etc/sections.cfg */
   mk_filename(str, DFL_SECTIONFILE, DFL_SECTIONFILE);
   if ((file = fopen(str,"rt")) == NULL) { 
      errlog("fopen(%s): %s (%i)",str,strerror(errno),errno);
      return 1;
   }        
   while (readline(str,sizeof(str),file)) {
      i = parse_section_config_line(str);
      if (i < 0) {
         errlog("section_read_config(): syntax error (%i): \"%s\"",i,str);
      }
      else if (i > 0){
	 s = section + section_count - 1;
	 
	 found_stat = 0; found_config = 0;
	 for (i = 0; (i < section_count - 1); i++) {
            if (!strcasecmp(s->config_file, (section + i)->config_file)) {
               s->config = (section + i)->config;
               s->config_size = (section + i)->config_size;
               s->config_count = (section + i)->config_count;
               found_config = 1; /* same configfile used multiple times */
            }
            if (!strcasecmp(s->stat_file, (section + i)->stat_file)) {
               s->stat = (section + i)->stat;
               s->stat_size = (section + i)->stat_size;
               s->stat_count = (section + i)->stat_count;
               found_stat = 1; /* same statfile used multiple times */
            }
	 }

         /* config file was not loaded yet */
	 if (!found_config) {
             s->config_size = s->config_count = 0;
	     s->config = NULL;
	     if (load_config_for_single_section(s) == -1) {
        	errlog("Error reading %s, using default one",s->config_file);
                s->config = section->config;
                s->config_size = section->config_size;
                s->config_count = section->config_count;
             }
	 }   
         /* stat file was not loaded yet */
	 if (!found_stat) {
             s->stat_size = s->stat_count = 0;
	     s->stat = NULL;
	     if (load_stat_file(s) == -1) {
        	errlog("File %s does not exist, creating new one",s->stat_file);
                if (init_stat_file(s) == -1) {
                   errlog("Error creating default stat file");
	           return -1;
                }
                save_stat_file(s);
             }
	 }   
      }
   }
   fclose(file);
   sort_sections(); /* Sort sections by dir length */
   return 0;
}


int alloc_section(void)
{ 
   void *ptr = NULL;

   ptr = g_realloc(section, (section_size + 1) * sizeof(struct section_s));
   if (!ptr) return section_size;
   section = ptr;
   section_size++;
   return section_size;
}


int parse_section_config_line(char* str)
{ 
   char*             q;
   struct section_s* s;

   if (section_count >= section_size) { 
      alloc_section();
      if (section_count >= section_size) { 
         errlog("out of memory (sections)");
         return 0;
      }
   }     
   s = section + section_count;
   q = str;
  
   /* scan section name */
   while (isspace(*q)) { q++; str++; }
   if ((*q == '#') || (*q == ';')) return 0;

   if (skipnspace(&q)) {
      if (q - str >= sizeof(s->section)) return -1;
      strncpy(s->section, str, q - str);
      *(s->section + (q - str)) = '\0';
      if (!skipspace(&q)) return -1;
      str = q;      
   } 
   else return -1;
  
   skipnspace(&q); 
   if (q - str >= sizeof(s->config_file)) return -1;
   strncpy(s->config_file, str, q-str);
   *(s->config_file + (q-str)) = '\0';
   if (!skipspace(&q)) return -1;
   str = q;
   
   skipnspace(&q); 
   if (q - str >= sizeof(s->stat_file)) return -1;
   strncpy(s->stat_file, str, q-str);
   *(s->stat_file + (q-str)) = '\0';
   if (!skipspace(&q)) return -1;

   if (strlen(q) >= sizeof(s->path)) return -1;
   strncpy(s->path, q, sizeof(s->path) - 1);
   if (*(s->path + strlen(s->path) - 1) == '\n') {
      *(s->path+strlen(s->path)-1) = 0;
   }
   /* for lame users that left white spaces at the end of line */
   q = s->path + strlen(s->path) - 1;
   while (*q == ' ') {
      *q = 0;
      q--;
   }
   /* all ok */
   section_count++;
   return 1;
}


void sort_sections()
{ 
   struct section_s  sectmp;
   struct section_s  *sec;
   int               n;
   int               sorted = 1;
  
   sec = &sectmp;
   while (sorted) {
      sorted = 0;
      for (n = 0; n < section_count - 1; n++) {
         if (strlen((section + n)->path) < strlen((section + n + 1)->path)) {
            memmove(sec, (section+n), sizeof(struct section_s));
            memmove((section + n), (section + n + 1), sizeof(struct section_s));
            memmove((section + n + 1), sec, sizeof(struct section_s));
            sorted = 1;
         }
      }
   } 
}


int alloc_config_key(struct section_s* s)
{ 
   int   new_size; 
   void  *ptr;

   new_size = s->config_size + 1;
   ptr = g_realloc(s->config, new_size * sizeof(struct dircfg_s));
   if (!ptr) {
      return s->config_size;
   }
   s->config = ptr;
   s->config_size = new_size;
   return s->config_size;       
}


int parse_config_for_section_line(char* str, struct section_s* s)
{ 
   char*             q;
   struct dircfg_s*  l;
   int               type;  
   int               i;

   if (s->config_count >= s->config_size) { 
      alloc_config_key(s);
      if (s->config_count >= s->config_size) { 
         errlog("out of memory (dircfg)");
         return -1;
      }
   }
   l = s->config + s->config_count;
   l->key = NULL;
   q = str;

   while (isspace(*q)) { 
      q++; 
      str++; 
   }
   if ((*q == '#') || (*q == ';')) 
      return 0;
   if (skipnspace(&q)) {
      *(q++) = 0;
      for (i = 0; i < (sizeof(key_type)/sizeof(struct key_type_s)); i++) {
         if (!strcasecmp(str, key_type[i].key)) {
            l->key = key_type[i].key;
            break;
         }
      }
      if (l->key) {
         type = get_key_type(l);
         switch (type) { 
            case D_BOOL : if (-1 == get_bool_key(l, q)) return -1; break; 
            case D_CHAR : if (-1 == get_char_key(l, q)) return -1; break;
            case D_HEX  : 
            case D_INT  : if (-1 == get_int_key(l, q)) return -1;  break; 
            default     : return 0; /* q is not a valid key */
         }
         s->config_count++; 
         return 1;
      }
   }
   return 0;
}


int get_key_type(struct dircfg_s* l) /* do binary search of key in key table */
{ 
   int   i, k, j, r;
   char *key;

   key = l->key;
   i = 0;
   k = (sizeof(key_type) / sizeof(struct key_type_s)) -1;
   while (i <= k) {
      j = (i + k);
      r = strcasecmp(key, key_type[j].key);
      if (r > 0) i = j + 1;
      else if (r < 0) k = j -1;
      else {
         l->type=key_type[j].value;
         return(key_type[j].value);
      }
   }  
   return 0; /* key not found in table */
}


int get_bool_key(struct dircfg_s*  l, char* q)
{ 
  while (isspace(*q)) { q++; }
  
  if (!*q) return -1;

  if (!strcasecmp(q, "yes") || !strcasecmp(q, "on") || !strcasecmp(q, "1") || !strcasecmp(q, "true") || !strcasecmp(q, "enabled"))
    l->value = (void *) YES;
  else
    l->value = (void *) NO;

  return 0;
}

int get_char_key(struct dircfg_s*  l, char* q)
{  
  while (isspace(*q)) q++; 
  if (*q) {
    if (*(q+strlen(q)-1) == '\n') *(q+strlen(q)-1) = 0;
    l->value = g_malloc(strlen(q) + 1);
    if(!l->value) return -1; 
    strcpy((char*)l->value, q); 
    return 0;
  }
  else { 
    l->value = g_malloc(sizeof(char));
    strcpy(l->value,"");
    return 0;
  }
    return -1;
}


int get_int_key(struct dircfg_s*  l, char* q)
{ int n;

  while (isspace(*q)) q++;
  if (*q) {
    if(1 != sscanf(q,"%i",&n)) return -1;
    /* l->value = g_malloc(sizeof(int)); fixme: WTF is that meant for -- memleaking on purpose !?!?!? */
    /* if(!l->value) return -1; */
    l->value = GINT_TO_POINTER (n);
    return 0;
  }
  return -1;   
} 

struct section_s* get_section_by_path(char *str)
{ 
   struct section_s *s;  
   int i;
  
   if (*(str+strlen(str)-1) == '\n') *(str+strlen(str)-1) = 0; 
   for (i = 0; i < section_count; i++) {
      s = section + i;
      if (!strncmp(str, s->path, strlen(s->path))) return s;
   }
   return NULL;
}
            
struct section_s* get_section_by_id(int id)
{
   struct section_s *s;
   int i;
   int x;

   for (x = 0, i = section_count; i > 0; i--, x++) {
      if ( x == id) {
         s = section + i - 1;
         return s;
      }
   }
   return NULL;
}

struct section_s* get_section_by_name(char *str)
{
   struct section_s *s;
   int i;
   
   if (*(str+strlen(str)-1) == '\n') *(str+strlen(str)-1) = 0; 
   for (i = 0; i < section_count; i++) {
      s = section + i;
      if (!strcasecmp(s->section,str)) return s;
   }
   return NULL;
}

struct sstat_s* get_stat_by_uid(struct section_s *s, gint32 uid)
{ 
   int i;
   struct sstat_s *st;
   
   if(!s || !(s->stat)) return NULL;
   st = s->stat;   
   
   for (i = 0; i < s->stat_count; i++) {
      if ((st+i)->uid == uid) {
	 return (st+i);
      }
   }
   return NULL;
}


int get_section_id(struct child_s* c, char** p, int n)
{ 
   struct section_s *s;
   char             str[256] = "";
   int              x;
   int              i;

   g_snprintf(str, sizeof(str), "%s ", p[0]);
   for(x = 1; x < n; x++) {
      strlcat(str, p[x], sizeof(str) - strlen(str));
      strlcat(str, " ", sizeof(str) - strlen(str));
   }
   *(str + strlen(str) - 1) = '\0';
   s = get_section_by_path(str);
  
   memset(str, 0, sizeof(str));
   if (s) {
      for (x = 0, i = section_count; i > 0; i--, x++) {
          if (s == section + i - 1)
             g_snprintf(str, sizeof(str), "+%d", x);
      }
      if (*str == '\0')
         strlcpy(str, "-Error searching section", sizeof(str));
   }
   else { 
      strlcpy(str, "-Error searching section", sizeof(str));
   }
   server_reply(c,str);
   return 0;
}

int get_section_name(struct child_s* c, char** p, int n)
{ 
   struct section_s *s;
   char             str[256];
   int              x;

   g_snprintf(str, sizeof(str), "%s ", p[0]);
   for(x = 1; x < n; x++) {
      strlcat(str, p[x], sizeof(str));
      strlcat(str, " ", sizeof(str));
   }
   *(str + strlen(str) - 1) = '\0';
   s = get_section_by_path(str);
  
   memset(str, 0, sizeof(str));
   if (s) g_snprintf(str, sizeof(str), "+%s", s->section);
   else strlcpy(str, "-Error on searching section", sizeof(str));
   server_reply(c,str);
   return 0;
}

int get_section_credits(struct child_s* c, char** p, int n)
{
   struct section_s *s;
   char             str[256];
   int              x;
   struct sstat_s   *st;

   g_snprintf(str, sizeof(str), "%s ", p[0]);
   for (x = 1; x < n; x++) {
      strlcat(str, p[x], sizeof(str));
      strlcat(str, " ", sizeof(str) - strlen(str));
   }
   *(str + strlen(str) - 1) = '\0';
   s = get_section_by_path(str);
   memset(str, 0, sizeof(str));
   if (s) {
      st = get_stat_by_uid(s, c->user->uid);
      if (st) g_snprintf(str, sizeof(str), "+" SI64, st->credit);
      else g_snprintf(str, sizeof(str), "-Error getting section credit");
   }
   else strlcpy(str, "-Error searching section", sizeof(str));
   server_reply(c,str);
   return 0;
}

/*  p[0] = key, p[1] - p[99] = directory */
int get_value_by_dir(struct child_s* c, char** p, int n)
{ 
   struct section_s* s;
   struct dircfg_s*  l;
   char              str[256];
   int               x;

   g_snprintf(str, sizeof(str), "%s ", p[1]);
   for(x = 2; x < n; x++) {
     strlcat(str, p[x], sizeof(str));
     strlcat(str, " ", sizeof(str));
   }
   *(str+strlen(str)-1) = 0;
   if (*(str+strlen(str)-1) == '\n') *(str+strlen(str)-1) = 0;
   s = get_section_by_path(str);
   if (search_key(p[0], s, &l)) { /* we get default section then */
      s = get_section_by_name("default");
      if (search_key(p[0], s, &l)) {
         g_snprintf(str, sizeof(str), "-key %s not found\n", p[0]); /* this cant happen */
         server_reply(c,str);
         return 0;
      }
   }
   if(l->type == D_BOOL) g_snprintf(str, sizeof(str), "+%i\n", (l->value == 0) ? 0 : 1);
   else if (l->type == D_INT || l->type == D_HEX) g_snprintf(str, sizeof(str), "+%i\n", GPOINTER_TO_INT (l->value)); 
   else g_snprintf(str, sizeof(str), "+%s\n", (char *)l->value);

   server_reply(c,str);
   return(0);
} 

int search_key(char *key, struct section_s *s, struct dircfg_s** l)  
{ 
   struct dircfg_s*   array = s->config;
   int                x = 0;

   if (!s) return 0;
   while(x < s->config_count) {
      if(!strcmp(array->key, key)) {
         if (l != NULL) *l = array;
         return 0;
      }
      array++;
      ++x;
   }
   return 1;
}


/* header for the stat files (mainly for version check) */
struct sf_ident_s
{ 
   char id[32];
   int  version;
   int  count;
   int  temp1;
   int  temp2;
} s_id = { "OpenFTPD stat file", 2, 0, 0, 0 };


int init_stat_file(struct section_s *s) 
{
   int            i;
   struct user_s* u;
   struct sstat_s*   new_stat;

   new_stat = (struct sstat_s*) g_realloc(s->stat, user_count * 
               sizeof(struct sstat_s));

   if(new_stat) { 
      memset(new_stat, 0, user_count * sizeof(struct sstat_s)); 
      for (i = 0, u = user; i < user_count; i++, u++) {
         (new_stat + i)->uid = u->uid;
      }
      s->stat = new_stat;
      s->stat_count = user_count;
   }
   return 0;
}


/* read dircfg files */
int load_config_for_single_section(struct section_s* s) 
{ 
   FILE* file;
   char * str;
   char  filename[256];
   int   i;

   str = (char *) g_malloc(BUFFER_LEN * sizeof(char));
   g_snprintf(filename, sizeof(filename), "%s%s", DFL_SECTIONDIR, s->config_file);
   mk_filename(str, filename, NULL);

   if ((file = fopen(str,"rt")) == NULL) { 
      errlog("fopen(%s): %s (%i)",str,strerror(errno),errno);
      return -1;
   }
   while (readline(str, BUFFER_LEN, file)) { 
      i = parse_config_for_section_line(str, s);
      if (i < 0) {
         errlog("load_sectionfile(): syntax error (%i): \"%s\" (%s)", 
	        i, str, s->config_file);
      }
      g_free(str);
      str = (char *) g_malloc(BUFFER_LEN * sizeof(char));
   }
   g_free(str);
   fclose(file);
   return 0;
} 

/* return  
   bool key: 0 if no, 1 if yes
   int  key: asigned to int* v
   char key: asigned to char *str 
   -1 if fail (should not happen) */
int get_key_value_by_section( struct section_s* s, char *key, char *str, size_t str_len, int *v)
{
   struct dircfg_s *l;
   int i;

   for (i = 0; i < s->config_count; i++) {
      l = s->config + i;   
      if (!strcasecmp(key,l->key)) {
         if (l->type == D_BOOL) return (l->value == NULL) ? FALSE : TRUE;
         else if (l->type == D_INT || l->type == D_HEX) *v = GPOINTER_TO_INT (l->value);
         else strlcpy(str,(char*)l->value, str_len);
         return 0;
      }
   } /* key not found in section, we call default one */
   s = get_section_by_name("default");
   if (s) {
      for (i = 0; i < s->config_count; i++) {
	 l = s->config + i;   
	 if (!strcasecmp(key,l->key)) {
	    if (l->type == D_BOOL) return (l->value == NULL) ? FALSE : TRUE;
            else if (l->type == D_INT || l->type == D_HEX) *v = GPOINTER_TO_INT (l->value);
            else strlcpy(str,(char*)l->value, str_len);
            return 0;
	 }
      }
   }
   return -1; 
}

int load_stat_file_v3012(struct section_s* s)
{ 
   FILE *                  file;
   char                    filename[256];
   char                    str[BUFFER_LEN];
   struct sf_ident_s       id_test;
   struct sstat_v3012_601* old_stat;
   struct sstat_v3012_601* os;
   struct sstat_s*         new_stat;
   int                     i;
   int                     n;

   if (*(stat_dir + strlen(stat_dir) - 1) == '/') 
      g_snprintf(filename, sizeof(filename), "%s%s", stat_dir, s->stat_file);
   else
      g_snprintf(filename, sizeof(filename), "%s/%s", stat_dir, s->stat_file);
   mk_filename(str, filename, NULL);

   file = fopen(str, "rb");
   if (!file) {
      errlog("fopen(%s): %s (%i)", str, strerror(errno), errno);
      return -1;
   }
   if (!fread(&id_test, sizeof(id_test), 1, file)) {
      errlog("fread(stat file): %s (%i)",strerror(errno),errno);
      return -1;
   }
   n = id_test.count;
   old_stat = (struct sstat_v3012_601*) g_realloc(NULL, n*sizeof(struct sstat_v3012_601));

   if (!old_stat) {
      errlog("out of memory (stats)");
      fclose(file);
      return -1;
   }

   if (fread(old_stat,sizeof(struct sstat_v3012_601),n,file) != n) {
      errlog("fread(stat file(v3012_601)): %s (%i)",strerror(errno),errno);
      fclose(file);
      s->stat_count = 0;
      return -1;
   } 
   s->stat = (struct sstat_s*)g_realloc(s->stat,n*sizeof(struct sstat_s));
   if (!s->stat) {
      errlog("out of memory (stats)");
      fclose(file);
      return -1;
   }
   memset(s->stat,0,n*sizeof(struct sstat_s));
   new_stat = s->stat;
   os = old_stat;
   for (i = 0; i < n; i++) {
      memcpy(new_stat++,old_stat++,sizeof(struct sstat_v3012_601));
   }
   s->stat_count = n;

   g_free(os);
   fclose(file);
   section_stat_dirty = 1;
   return 0;
}

int load_stat_file_v1beta29(struct section_s* s)
{ 
   FILE *                  file;
   char                    filename[256];
   char                    str[BUFFER_LEN];
   struct sf_ident_s       id_test;
   struct sstat_v1_beta29* old_stat;
   struct sstat_v1_beta29* os;
   struct sstat_s*         new_stat;
   int                     i;
   int                     n;

   if (*(stat_dir + strlen(stat_dir) - 1) == '/') 
      g_snprintf(filename, sizeof(filename), "%s%s", stat_dir, s->stat_file);
   else
      g_snprintf(filename, sizeof(filename), "%s/%s", stat_dir, s->stat_file);
   mk_filename(str, filename, NULL);

   file = fopen(str, "rb");
   if (!file) {
      errlog("fopen(%s): %s (%i)", str, strerror(errno), errno);
      return -1;
   }
   if (!fread(&id_test, sizeof(id_test), 1, file)) {
      errlog("fread(stat file): %s (%i)",strerror(errno),errno);
      return -1;
   }
   n = id_test.count;
   old_stat = (struct sstat_v1_beta29*) g_realloc(NULL, n*sizeof(struct sstat_v1_beta29));

   if (!old_stat) {
      errlog("out of memory (stats)");
      fclose(file);
      return -1;
   }

   if (fread(old_stat,sizeof(struct sstat_v1_beta29),n,file) != n) {
      errlog("fread(stat file(v1 beta 29)): %s (%i)",strerror(errno),errno);
      fclose(file);
      s->stat_count = 0;
      return -1;
   } 
   s->stat = (struct sstat_s*)g_realloc(s->stat,n*sizeof(struct sstat_s));
   if (!s->stat) {
      errlog("out of memory (stats)");
      fclose(file);
      return -1;
   }
   memset(s->stat,0,n*sizeof(struct sstat_s));
   new_stat = s->stat;
   os = old_stat;
   for (i = 0; i < n; i++) {
      memcpy(new_stat++,old_stat++,sizeof(struct sstat_v1_beta29));
   }
   s->stat_count = n;

   g_free(os);
   fclose(file);
   section_stat_dirty = 1;
   return 0;
}

int load_stat_file(struct section_s* s)
{ 
   FILE*               file;
   char                filename[256];
   char                str[BUFFER_LEN];
   struct sf_ident_s   id_test;
   int                 n;
   int                 new_size;
   struct sstat_s      *new_stat;

   g_snprintf(filename, sizeof(filename), "%s%s", DFL_SECTIONSTATDIR, s->stat_file);
   mk_filename(str, filename, NULL);

   file = fopen(str, "rb");
   if (!file) { 
      errlog("fopen(%s): %s (%i)", str, strerror(errno), errno);
      return -1;
   }

   if (!fread(&id_test, sizeof(id_test), 1, file)) { 
      fclose(file);
      errlog("fread(stat file): %s (%i)",strerror(errno),errno);
      return -1;
   }
   if (strcmp(id_test.id, s_id.id) && strcmp(id_test.id, "FTP4ALL stat file")) { 
      errlog("stat file %s has bad ID", filename);
      fclose(file);
      return -1;
   } 
   if (id_test.version > s_id.version) { 
      errlog("stat file is newer version than daemon");
      fclose(file);
      return -1;
   }
   if (id_test.version < s_id.version) { 
      errlog("stat file is older version than daemon");
      fclose(file);
      if (id_test.version == 0) {
         errlog("stat file is 3.012-6.1 version");
         return load_stat_file_v3012(s);
      }
      else if (id_test.version == 1) {
         errlog("stat file is v1.beta29 version");
         return load_stat_file_v1beta29(s);
      }
      return -1;
   }
   
   n = id_test.count;            
   new_stat = g_realloc(s->stat, n * sizeof(struct sstat_s));
   if (!new_stat) {
      errlog("out of memory (stats)");
      fclose(file);
      return -1;
   }   
   new_size = fread(new_stat, sizeof(struct sstat_s), n, file);

   if (new_size != n) { 
      errlog("fread(stat file): %s (%i)",strerror(errno),errno);
      fclose(file);
      s->stat_count = 0;
      return -1;
   }

   s->stat = new_stat;
   s->stat_count = new_size;

   fclose(file);
   section_stat_dirty = 0;
   return 0;
}



/* save stat file of a section with index "section_index" to
   file "filename". number of records = user_count */

int save_stat_file(struct section_s *s)
{ 
   FILE*          file;
   char           fn_tmp[256];
   char           fn_tmp2[256];
   size_t         n;

   g_snprintf(fn_tmp, sizeof(fn_tmp), "%s%s/%s", server_dir, DFL_SECTIONSTATDIR, s->stat_file);
   g_snprintf(fn_tmp2, sizeof(fn_tmp2), "%s.tmp", fn_tmp);

   file = fopen(fn_tmp2, "wb");
   if (file == NULL) { 
      errlog("fopen(%s): %s (%i)", fn_tmp, strerror(errno), errno);
      return -1;
   }
   s_id.count = s->stat_count;

   if (!fwrite(&s_id, sizeof(s_id), 1 ,file)) { 
      errlog("fwrite(%s): %s (%i)", fn_tmp, strerror(errno), errno);
      fclose(file);
      return -1;
   }

   n = fwrite(s->stat, sizeof(struct sstat_s), s->stat_count, file);
   if (n != s->stat_count) { 
      errlog("fwrite(%s): %s (%i)", fn_tmp, strerror(errno), errno);
      fclose(file);
      return -1;
   } 
   else { 
      fclose(file);
      if (rename(fn_tmp2, fn_tmp) < 0) {
         errlog("rename(%s,%s): %s (%i)", fn_tmp, fn_tmp2, strerror(errno), errno);
         return -1;
      }
   }
   return 0;
}


int save_section_stat()
{
   int i;
   
   for (i = 0; i < section_count; i++) {
      if (section+i) {
         if(save_stat_file(section+i) == -1) return -1;
      }
   }
   return 0;
}


int write_text_stat_file(struct section_s *s)
{ 
   FILE*          file;
   char           filename[256];
   struct user_s  *u;
   struct group_s *g;
   struct sstat_s *st;
   int            i;
   char           str[128];

   strlcpy(str, s->section, sizeof(str));
   g_strdown (str);
   g_snprintf(filename, sizeof(filename), "%s%s%s%s.txt", server_dir, 
      stat_dir, (*(stat_dir + strlen(stat_dir) - 1) == '/') ? "" : "/", 
      str);

   file = fopen(filename, "wt");
   if (file == NULL) { 
      errlog("fopen(%s): %s (%i)", filename, strerror(errno), errno);
      return -1;
   }
   if (!(s->stat)) {
      errlog("s->stat is empty");
      return -1;
   }

   st = s->stat;

   for (i = 0; i < s->stat_count; i++) {
      u = get_user_by_uid((st+i)->uid);
      if(u) {
         g = get_group_by_gid(u->gid);
      }
      if (u && g) {
         if(!(u->class & UF_INVISIBLE)) {
	    fprintf(file, "%s\t%s\t%s\t%s\t"
                    SI32"\t"SI64"\t"SI64"\t"
                    SI32"\t"SI64"\t"SI64"\t"
                    SI32"\t"SI64"\t"SI64"\t"
                    SI32"\t"SI64"\t"SI64"\t"
                    SI32"\t"SI64"\t"SI64"\t"
                    SI32"\t"SI64"\t"SI64"\t"
                    SI32"\t"SI64"\t"SI64"\t"
                    SI32"\t"SI64"\t"SI64"\t"
                    SI32"\t"SI64"\t"SI64"\t"
                    SI32"\t"SI64"\t"SI64"\n",
                    u->name, u->info, g->name, g->info,
                    (st+i)->dl_f_day, (st+i)->dl_day, (st+i)->dl_tm_day, 
                    (st+i)->ul_f_day, (st+i)->ul_day, (st+i)->ul_tm_day, 
                    (st+i)->dl_f_week, (st+i)->dl_week, (st+i)->dl_tm_week, 
                    (st+i)->ul_f_week, (st+i)->ul_week, (st+i)->ul_tm_week, 
        	    (st+i)->dl_f_month, (st+i)->dl_month, (st+i)->dl_tm_month,
        	    (st+i)->ul_f_month, (st+i)->ul_month, (st+i)->ul_tm_month,
                    (st+i)->dl_f_year, (st+i)->dl_year, (st+i)->dl_tm_year, 
                    (st+i)->ul_f_year, (st+i)->ul_year, (st+i)->ul_tm_year, 
        	    (st+i)->dl_f_total, (st+i)->dl_total, (st+i)->dl_tm_total,
        	    (st+i)->ul_f_total, (st+i)->ul_total, (st+i)->ul_tm_total);
         }
      }
   }
   fclose(file);
   return 0;
}

int write_global_stat_file()
{ 
   FILE*          file;
   char           filename[256];
   struct user_s  *u = NULL;
   struct group_s *g = NULL;
   int            i;

   if (*(stat_dir + strlen(stat_dir) - 1) == '/')
      g_snprintf(filename, sizeof(filename), "%s%sglobal.txt", server_dir, stat_dir);
   else
      g_snprintf(filename, sizeof(filename), "%s%s/global.txt", server_dir, stat_dir);

   file = fopen(filename, "wt");
   if (file == NULL) { 
      errlog("fopen(%s): %s (%i)", filename, strerror(errno), errno);
      return -1;
   }

   for (i = 0; i < user_count; i++) {
      u = user+i;
      if (u) g = get_group_by_gid(u->gid);
      if (g) {
	 fprintf(file, "%s\t%s\t%s\t%s\t"
                 SI32"\t"SI64"\t"SI64"\t"SI32"\t"
                 SI64"\t"SI64"\t"SI32"\t"SI64"\t"
                 SI64"\t"SI32"\t"SI64"\t"SI64"\t"
                 SI32"\t"SI64"\t"SI64"\t"SI32"\t"
                 SI64"\t"SI64"\t"SI32"\t"SI64"\t"
                 SI64"\t"SI32"\t"SI64"\t"SI64"\t"
                 SI32"\t"SI64"\t"SI64"\t"SI32"\t"
                 SI64"\t"SI64"\n",
                 u->name, u->info, g->name, g->info,
                 u->dl_f_day, u->dl_day, u->dl_tm_day, u->ul_f_day, 
                 u->ul_day, u->ul_tm_day, u->dl_f_week, u->dl_week,
                 u->dl_tm_week, u->ul_f_week, u->ul_week, u->ul_tm_week,
        	 u->dl_f_month, u->dl_month, u->dl_tm_month, u->ul_f_month,
        	 u->ul_month, u->ul_tm_month, u->dl_f_year, u->dl_year,
                 u->dl_tm_year, u->ul_f_year, u->ul_year, u->ul_tm_year,
        	 u->dl_f_total, u->dl_total, u->dl_tm_total, u->ul_f_total,
        	 u->ul_total, u->ul_tm_total);
      }
   }

   fclose(file);
   return 0;
}


int write_section_stat_files()
{
   int i;
   struct section_s *s;
   
   for (i = 0, s = section; i < section_count; i++, s++) {
      if(s) write_text_stat_file(s);
   }
   return 0;
}

void printf_section_keys(struct child_s* c, struct section_s* s, int id)
{   
    int i;
    int j;
    char str[256];
    char str2[sizeof(key_type[0].def)];
    int x;

    g_snprintf(str, sizeof(str), "+ !C| %70s |!0", " ");
    server_reply(c, str);
    g_snprintf(str, sizeof(str), "+ !C| SECTION: !H%-23s !CID:!H %-34d!C|!0", s->section, id);
    server_reply(c, str);
    g_snprintf(str, sizeof(str), "+ !C| PATH: !H%-64s !C|!0", s->path);
    server_reply(c, str);
    g_snprintf(str, sizeof(str), "+ !C| %70s |!0", " ");
    server_reply(c, str);
    
    for (i = 0; i < (sizeof(key_type)/sizeof(struct key_type_s)); i++) {
       if ((j = get_key_value_by_section(s, key_type[i].key, str2, sizeof(str2), &x)) != -1) {
          if (key_type[i].value == D_BOOL) 
             g_snprintf(str, sizeof(str), "+ !C| !H%-32s !0%-37s !C|!0\n", key_type[i].key, j ? "yes" : "no" );
          else if (key_type[i].value == D_INT)
             g_snprintf(str, sizeof(str), "+ !C| !H%-32s !0%-37d !C|!0\n", key_type[i].key, x);
          else if (key_type[i].value == D_HEX)
             g_snprintf(str, sizeof(str), "+ !C| !H%-32s !00x%-35X !C|!0\n", key_type[i].key, x);
          else
             g_snprintf(str, sizeof(str), "+ !C| !H%-32s !0%-37.37s !C|!0\n", key_type[i].key, str2);
          server_reply(c, str);
       }
    }
    g_snprintf(str, sizeof(str), "+ !C| %70s |!0", " ");
    server_reply(c, str);
}

void printf_section_names(struct child_s* c, struct section_s* s, int id)
{   
    char str[256];
    char str2[sizeof(key_type[0].def)];

    g_snprintf(str, sizeof(str), "+ | ID:!H %-2d !C| SECTION: !H%-10s !C |PATH: !H%-33s | !0", id, s->section, s->path);
    server_reply(c, str);
}

int f4adp_sec_list(struct child_s* c, char** p, int n)
{   
    struct section_s* s = NULL;
    int x;
    int id;

    if (n) {
       if (1 != sscanf(p[0], "%d", &id)) {
          server_reply(c, "-Wrong Section id");
          return 0;
       } 
       s = get_section_by_id(id);
       if (!s) {
          server_reply(c,"-Section does not exist");
          return 0;
       }
    } 
    server_reply(c, "+ !C.------------------------------------------------------------------------.!0");
    server_reply(c, "+ !C|!3 !BOpenFTPD!H SECTIONS KEYS                                                 !0!C|!0");
    server_reply(c, "+ !C|------------------------------------------------------------------------|!0");
    if (s)
       printf_section_keys(c, s, id);
    else {
       for (id = 0, x = section_count; x > 0; x--, id++) {
	  printf_section_names(c, section + x - 1, id);
//          printf_section_keys(c, section + x - 1, id);
       }
    }
    server_reply(c, "+ !C|------------------------------------------------------------------------|!0");
    server_reply(c, "+ !C|!H WARNING: Changing a default section key will change all sections that !C |!0");
    server_reply(c, "+ !C|!H didn't have this key defined in their config file                      !C|!0");
    server_reply(c, "+ !C`------------------------------------------------------------------------+!0");
    server_reply(c, "+");
    return 0;
}

int f4adp_sec_set(struct child_s* c, char** p, int n)
{
    struct section_s* s = NULL;
    struct dircfg_s*  l;
    int               x;
    int               found;
    int               id;
    char              str[BUFFER_LEN];
    char              filename[256];
    char              filenametmp[256];
    char              *q;
    FILE*             file;
    FILE*             filetmp;

    if (1 != sscanf(p[0], "%d", &id)) {
       server_reply(c, "-Section id is wrong");
       return 0;
    }
    s = get_section_by_id(id);
    if (!s) {
       server_reply(c, "-Section does not exist");
       return 0;
    }
    /* First del the old value (cause it can be not defined)*/
    for (x = 0; x < s->config_count; x++) {
        l = s->config + x;
        if (!strcasecmp(l->key, p[1])) { /* it will find the old one */
           memmove(l, l + 1, (--s->config_count-(l-s->config))*sizeof(struct dircfg_s));
           break;
        }
    }
    /* insert new value */
    g_snprintf(str, sizeof(str), "%s %s\n", p[1], p[2]);
    if (1 != parse_config_for_section_line(str, s)) {
       server_reply(c,"-Syntax error (key value)");
       return 0;
    }
   /* We write into the section.cfg file */
   g_snprintf(str, sizeof(str), "%s%s", DFL_SECTIONDIR, s->config_file);
   mk_filename(filename, str, NULL);
   g_snprintf(filenametmp, sizeof(filenametmp), "%s.tmp", filename);
   if ((file = fopen(filename,"rt")) == NULL) { 
      errlog("(f4adp_sec_set) fopen(%s): %s (%i)",str,strerror(errno),errno);
      server_reply(c, "-Key value changed temporary (Couldn't open config)");
      return -1;
   }
   if ((filetmp = fopen(filenametmp,"wt")) == NULL) {
      errlog("(f4adp_sec_set) fopen(%s): %s (%i)",str,strerror(errno),errno);
      server_reply(c, "-Key value changed temporary (Couldn't open config)");
      fclose(file);
      return -1;
   }
   found = 0;
   while (fgets(str, sizeof(str), file)) { 
      q = str;
      while (isspace(*q)) 
         q++;
      if (!strncasecmp(q, p[1], strlen(p[1])) && *(q + strlen(p[1])) == ' ') {
         found = 1;
         g_snprintf(str, sizeof(str), "%-20s               %s\n", p[1], p[2]);
      }
      fprintf(filetmp, str);
   }
   if (!found) {/* the section have this value from default one */
      fprintf(filetmp, "%-20s               %s\n", p[1], p[2]);
   }
   fclose(file);
   fclose(filetmp);
   remove(filename);
   rename(filenametmp, filename);
   server_reply(c, "+Key value changed");
   return 0;
}

int f4adp_section_change(struct child_s* c, char** p, int n)
{
   struct sstat_s*    st = NULL;
   struct section_s*  s = NULL;
   struct user_s*     u = NULL; 

   gint64              l;     /* integer (aux)           */
   gint32              i;     /* integer (aux)           */
   int                pd;    /* flag: permission denied */
   char*              q;     /* argument after "="      */
   int                se;    /* flag: syntax error      */
   char               str[256]; /* Aux buffer */
   int                dc;    /* flag: no different credit */
   int                id;

   if (1 != sscanf(p[0], "%d", &id)) {
      server_reply(c, "-Wrong Section id");
      return 0;
   } 
   s = get_section_by_id(id);
   if (!s) {
      server_reply(c,"-Section does not exist");
      return 0;
   }
   u = get_user_by_name(p[1]);
   if (!u) {
      server_reply(c,"-User does not exist");
      return 0;
   }
   st = get_stat_by_uid(s, u->uid);
   if (!st) {
      g_snprintf(str, sizeof(str), "-Could not find stats for user %s in section %s", 
              u->name, s->section);
      server_reply(c, str);
      errlog(str);
      return 0;
   }
   if (!allowed(c->user,R_USER_EDIT,u)) {
      server_reply(c, "-Permission denied.");
      return 0;
   }
   pd = se = dc = 0;
   p++;
   n--;
   while (--n) {
      p++;
      q = strchr(*p,'=');
      if (q) {
         *(q++) = '\0';
         if (!strcasecmp(*p,"CREDIT")) {
            if (!allowed(c->user, R_USER_EDIT_CREDIT, u)) pd = 1; 
            else if (!(get_key_value_by_section(s, "different_credit", NULL, 0,
                     NULL))) dc = 1;
            else if (!sscanf(q, SI64, &st->credit)) se = 1;
         }
         else if (!strcasecmp(*p,"CREDIT+")) { 
            if (!allowed(c->user, R_USER_EDIT_CREDIT, u)) pd = 1; 
            else if (!(get_key_value_by_section(s, "different_credit", NULL, 0,
                     NULL))) dc = 1;
            else if (sscanf(q, SI64, &l)) st->credit += l;
            else  se = 1;
         }
         else if (!strcasecmp(*p,"CREDIT-")) { 
            if (!allowed(c->user, R_USER_EDIT_CREDIT, u)) pd = 1; 
            else if (!(get_key_value_by_section(s, "different_credit", NULL, 0,
                     NULL))) dc = 1;
            else if (sscanf(q, SI64, &l)) st->credit -= l;
            else  se = 1;
         }
         else if (!strcasecmp(*p,"MBCREDIT")) { 
            if (!allowed(c->user, R_USER_EDIT_CREDIT, u)) pd = 1; 
            else if (!(get_key_value_by_section(s, "different_credit", NULL, 0,
                     NULL))) dc = 1;
            else if (sscanf(q, SI64, &l)) st->credit = l * 0x100000;
            else  se = 1;
         }
         else if (!strcasecmp(*p,"MBCREDIT+")) { 
            if (!allowed(c->user, R_USER_EDIT_CREDIT, u)) pd = 1; 
            else if (!(get_key_value_by_section(s, "different_credit", NULL, 0,
                     NULL))) dc = 1;
            else if (sscanf(q, SI64, &l)) st->credit += l * 0x100000;
            else  se = 1;
         }
         else if (!strcasecmp(*p,"MBCREDIT-")) { 
            if (!allowed(c->user, R_USER_EDIT_CREDIT, u)) pd = 1; 
            else if (!(get_key_value_by_section(s, "different_credit", NULL, 0,
                     NULL))) dc = 1;
            else if (sscanf(q, SI64, &l)) st->credit -= l * 0x100000;
            else  se = 1;
         }
         /* WEEKLY ALOTMENT */
         else if (!strcasecmp(*p, "WKALOT")) {
            if (!allowed(c->user, R_USER_EDIT_CREDIT, u)) pd = 1;
            else if (!(get_key_value_by_section(s, "different_credit", NULL, 0,
                     NULL))) dc = 1;
            else if (sscanf(q, SI64, &st->wkly_alot))
                 st->credit = st->wkly_alot;
            else se = 1;
         }
         else if (!strcasecmp(*p, "MBWKALOT")) {
            if (!allowed(c->user, R_USER_EDIT_CREDIT, u)) pd = 1;
            else if (!(get_key_value_by_section(s, "different_credit", NULL, 0,
                     NULL))) dc = 1;
            else if (sscanf(q, SI64, &l)) {
               st->wkly_alot = l * 0x100000;
               st->credit = st->wkly_alot;
            }
            else se = 1;
         }
         else if (!strcasecmp(*p,"UP")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->ul_total = l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"UP+")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->ul_total += l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"UP-")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->ul_total -= l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"DOWN")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->dl_total = l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"DOWN+")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->dl_total += l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"DOWN-")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->dl_total -= l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"DUP")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->ul_day = l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"DUP+")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->ul_day += l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"DUP-")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->ul_day -= l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"DDOWN")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->dl_day = l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"DDOWN+")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->dl_day += l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"DDOWN-")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->dl_day -= l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"WKUP")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->ul_week = l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"WKUP+")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->ul_week += l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"WKUP-")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->ul_week -= l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"WKDOWN")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->dl_week = l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"WKDOWN+")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->dl_week += l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"WKDOWN-")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->dl_week -= l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"MUP")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->ul_month = l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"MUP+")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->ul_month += l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"MUP-")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->ul_month -= l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"MDOWN")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->dl_month = l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"MDOWN+")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->dl_month += l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"MDOWN-")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->dl_month -= l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"YUP")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->ul_year = l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"YUP+")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->ul_year += l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"YUP-")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->ul_year -= l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"YDOWN")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->dl_year = l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"YDOWN+")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->dl_year += l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"YDOWN-")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI64, &l)) st->dl_year -= l * 0x100000;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FUP")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->ul_f_total = i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FUP+")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->ul_f_total += i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FUP-")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->ul_f_total -= i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FDOWN")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->dl_f_total = i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FDOWN+")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->dl_f_total += i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FDOWN-")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->dl_f_total -= i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FDUP")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->ul_f_day = i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FDUP+")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->ul_f_day += i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FDUP-")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->ul_f_day -= i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FDDOWN")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->dl_f_day = i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FDDOWN+")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->dl_f_day += i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FDDOWN-")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->dl_f_day -= i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FWKUP")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->ul_f_week = i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FWKUP+")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->ul_f_week += i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FWKUP-")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->ul_f_week -= i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FWKDOWN")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->dl_f_week = i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FWKDOWN+")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->dl_f_week += i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FWKDOWN-")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->dl_f_week -= i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FMUP")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->ul_f_month = i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FMUP+")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->ul_f_month += i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FMUP-")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->ul_f_month -= i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FMDOWN")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->dl_f_month = i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FMDOWN+")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->dl_f_month += i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FMDOWN-")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->dl_f_month -= i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FYUP")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->ul_f_year = i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FYUP+")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->ul_f_year += i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FYUP-")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->ul_f_year -= i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FYDOWN")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->dl_f_year = i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FYDOWN+")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->dl_f_year += i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"FYDOWN-")) {
            if (!allowed(c->user, R_USER_EDIT_STATS, u)) pd = 1; 
            else if (sscanf(q, SI32, &i)) st->dl_f_year -= i;
            else se = 1;
         }
         else if (!strcasecmp(*p,"KBSUP")) {
            if (!allowed(c->user, R_USER_EDIT_SPEEDS, u)) pd = 1; 
            else if (sscanf(q,SI64,&l)) st->ul_tm_total = (st->ul_total/1024)/l;
            else se = 1;
         }
         else if (!strcasecmp(*p,"KBSDOWN")) {
            if (!allowed(c->user, R_USER_EDIT_SPEEDS, u)) pd = 1; 
            else if (sscanf(q,SI64,&l)) st->dl_tm_total = (st->dl_total/1024)/l;
            else se = 1;
         }
         else if (!strcasecmp(*p,"KBSDUP")) {
            if (!allowed(c->user, R_USER_EDIT_SPEEDS, u)) pd = 1; 
            else if (sscanf(q,SI64,&l)) st->ul_tm_day = (st->ul_day/1024)/l;
            else se = 1;
         }
         else if (!strcasecmp(*p,"KBSDDOWN")) {
            if (!allowed(c->user, R_USER_EDIT_SPEEDS, u)) pd = 1; 
            else if (sscanf(q,SI64,&l)) st->dl_tm_day = (st->dl_day/1024)/l;
            else se = 1;
         }
         else if (!strcasecmp(*p,"KBSWKUP")) {
            if (!allowed(c->user, R_USER_EDIT_SPEEDS, u)) pd = 1; 
            else if (sscanf(q,SI64,&l)) st->ul_tm_week = (st->ul_week/1024)/l;
            else se = 1;
         }
         else if (!strcasecmp(*p,"KBSWKDOWN")) {
            if (!allowed(c->user, R_USER_EDIT_SPEEDS, u)) pd = 1; 
            else if (sscanf(q,SI64,&l)) st->dl_tm_week = (st->dl_week/1024)/l;
            else se = 1;
         }
         else if (!strcasecmp(*p,"KBSMUP")) {
            if (!allowed(c->user, R_USER_EDIT_SPEEDS, u)) pd = 1; 
            else if (sscanf(q,SI64,&l)) st->ul_tm_month = (st->ul_month/1024)/l;
            else se = 1;
         }
         else if (!strcasecmp(*p,"KBSMDOWN")) {
            if (!allowed(c->user, R_USER_EDIT_SPEEDS, u)) pd = 1; 
            else if (sscanf(q,SI64,&l)) st->dl_tm_month = (st->dl_month/1024)/l;
            else se = 1;
         }
         else if (!strcasecmp(*p,"KBSYUP")) {
            if (!allowed(c->user, R_USER_EDIT_SPEEDS, u)) pd = 1; 
            else if (sscanf(q,SI64,&l)) st->ul_tm_year = (st->ul_year/1024)/l;
            else se = 1;
         }
         else if (!strcasecmp(*p,"KBSYDOWN")) {
            if (!allowed(c->user, R_USER_EDIT_SPEEDS, u)) pd = 1; 
            else if (sscanf(q,SI64,&l)) st->dl_tm_year = (st->dl_year/1024)/l;
            else se = 1;
         }
         else se = 1;
      }
      else se = 1;
      if (pd || se || dc) n = 1;
   }
   if (pd) {
      server_reply(c,"-Permission denied.");
      return 0;
   } 
   else if (se) {
      server_reply(c,"-Syntax error.");
      return 0;
   } 
   else if (dc) {
      server_reply(c,"-Section doesn't have different credit.");
      return 0;
   } 
   section_stat_dirty++;
   server_reply(c,"+User changed.");
   return 0;
}


int f4adp_section_reset(struct child_s* c, char** p, int n)
{
   int               r_time = 0;
   int               r_what = 0;
   int               se     = 0;
   int               all    = 0;
   int               i;
   int               id;

   struct section_s* s      = NULL;
   struct sstat_s*   st     = NULL;
      
   if (strcasecmp(p[0],"all")) {
      if (1 != sscanf(p[0], "%d", &id)) {
         server_reply(c, "-Wrong Section id");
         return 0;
      } 
      s = get_section_by_id(id);
   }
   else {
      s = section;
      all = 1;
   }
   if (!s) {
      server_reply(c, "-Section does not exist");
      return 0;
   }

   if (!strcasecmp(p[1],"all")) r_what = 0x07; 
   else if (!strcasecmp(p[1],"up")) r_what = 0x01;
   else if (!strcasecmp(p[1],"down")) r_what = 0x02; 
   else if (!strcasecmp(p[1],"credit")) r_what = 0x04; 
   else se = 1;

   if (r_what & 0x03) {
      if (!strcasecmp(p[2],"all")) r_time = 0x01f;
      else if (!strcasecmp(p[2],"week")) r_time = 0x01;
      else if (!strcasecmp(p[2],"month")) r_time = 0x02;
      else if (!strcasecmp(p[2],"year")) r_time = 0x04;
      else if (!strcasecmp(p[2],"total")) r_time = 0x08;
      else if (!strcasecmp(p[2],"day")) r_time = 0x10;
      else se = 1;
      if (r_time != 0x01f) r_what &= 0x03;
   } 
   else se = 1;

   if (se) {
      server_reply(c, "-Syntax error");
      return 0;
   }
   if (all) {
      for (n = 0, s = section; n < section_count; n++, s++) {
          for (i = 0, st = s->stat; i < s->stat_count; i++, st++) {
             if (r_what & 0x01) {
                if (r_time & 0x01) 
                   st->ul_f_week = st->ul_week = st->ul_tm_week = 0;
                if (r_time & 0x02) 
                   st->ul_f_month = st->ul_month = st->ul_tm_month = 0;
                if (r_time & 0x04) 
                   st->ul_f_year = st->ul_year = st->ul_tm_year = 0;
                if (r_time & 0x08) 
                   st->ul_f_total = st->ul_total = st->ul_tm_total = 0;
                if (r_time & 0x10) 
                   st->ul_f_day = st->ul_day = st->ul_tm_day = 0;
             }
             if (r_what & 0x02) {
                if (r_time & 0x01) 
                   st->dl_f_week = st->dl_week = st->dl_tm_week = 0;
                if (r_time & 0x02) 
                   st->dl_f_month = st->dl_month = st->dl_tm_month = 0;
                if (r_time & 0x04) 
                   st->dl_f_year = st->dl_year = st->dl_tm_year = 0;
                if (r_time & 0x08) 
                   st->dl_f_total = st->dl_total = st->dl_tm_total = 0;
                if (r_time & 0x10) 
                   st->dl_f_day = st->dl_day = st->dl_tm_day = 0;
             }
             if (r_what & 0x04) {
                st->credit = 0;
             }
          }
      }
   }
   else { 
      for (n = 0, st = s->stat; n < s->stat_count; n++, st++) {
         if (r_what & 0x01) {
            if (r_time & 0x01) 
               st->ul_f_week = st->ul_week = st->ul_tm_week = 0;
            if (r_time & 0x02) 
               st->ul_f_month = st->ul_month = st->ul_tm_month = 0;
            if (r_time & 0x04) 
               st->ul_f_year = st->ul_year = st->ul_tm_year = 0;
            if (r_time & 0x08) 
               st->ul_f_total = st->ul_total = st->ul_tm_total = 0;
            if (r_time & 0x10) 
               st->ul_f_day = st->ul_day = st->ul_tm_day = 0;
         }
         if (r_what & 0x02) {
            if (r_time & 0x01) 
               st->dl_f_week = st->dl_week = st->dl_tm_week = 0;
            if (r_time & 0x02) 
               st->dl_f_month = st->dl_month = st->dl_tm_month = 0;
            if (r_time & 0x04) 
               st->dl_f_year = st->dl_year = st->dl_tm_year = 0;
            if (r_time & 0x08) 
               st->dl_f_total = st->dl_total = st->dl_tm_total = 0;
            if (r_time & 0x10) 
               st->dl_f_day = st->dl_day = st->dl_tm_day = 0;
         }
         if (r_what & 0x04) {
            st->credit = 0;
         }
      }
   }
   section_stat_dirty++;
   server_reply(c,"+Reset successful");
   return 0;

}

int f4adp_get_section_config(struct child_s* c, char** p, int n)
{ 
   struct section_s* s;
   char              str[256];
   int               x;

   memset(str, 0, sizeof(str));
   g_snprintf(str, sizeof(str), "%s ", p[0]);
   for(x = 1; x < n; x++) {
     strlcat(str, p[x], sizeof(str));
     strlcat(str, " ", sizeof(str));
   }
   if (*(str+strlen(str)-1) == '\n') *(str+strlen(str)-1) = 0;
   s = get_section_by_path(str);
   g_snprintf(str, sizeof(str), "+%s\n", s ? s->config_file : "default.cfg");
   server_reply(c,str);
   return(0);
} 
/* EOF */       
